import os
from collections.abc import Mapping
from typing import Final

import pandas as pd
import sqlalchemy

from .._literals import NumericalTypeHint
from ..base import AbstractConfig
from ..dataset import (
    AdmissionIntervalEventsTableColumns,
    AdmissionIntervalRatesTableColumns,
    AdmissionsTableColumns,
    COLUMN,
    MultivariateTimeSeriesTableMeta,
    StaticTableColumns,
    TableColumns,
)
from ..utils import resources_path
from .mimic import (
    CodedColumns,
    CodedTableResource,
    DatasetTablesResources,
    GroupedMultivariateTimeSeriesTableResource,
    MixedICDTableResource_MIMICIV,
    MultivariateTimeSeriesTableResource,
    StaticTableResource_MIMICIV,
    TableResource,
)


class SQLTableInterface(AbstractConfig):
    # resource file.
    query_template: str | None
    substitutes: Mapping[str, str]

    def __init__(self, query_template: str | None, substitutes: dict[str, str] | None = None):
        self.substitutes = substitutes or {}
        self.query_template = query_template

    @property
    def query(self) -> str:
        assert self.query_template is not None, "Query template must be set."
        return open(resources_path(self.query_template)).read()

    def load_standard_columns_table(self, engine: sqlalchemy.Engine):
        sub = COLUMN.as_dict() | dict(self.substitutes)
        query = self.query.format(**sub)
        return pd.read_sql(query, engine, coerce_float=False)


class SQLCodedTableInterface(SQLTableInterface):
    query_template: str | None
    space_query_template: str | None

    def __init__(
        self,
        query_template: str | None = None,
        space_query_template: str | None = None,
        substitutes: dict[str, str] | None = None,
    ):
        super().__init__(query_template=query_template, substitutes=substitutes)
        self.space_query_template = space_query_template

    @property
    def space_query(self) -> str:
        assert self.space_query_template is not None, "Space query template must be set."
        return open(resources_path(self.space_query_template)).read()

    def load_space_table(self, engine: sqlalchemy.Engine) -> pd.DataFrame:
        """
        Load the space table for the coded table.
        """
        query = self.space_query.format(**COLUMN.as_dict())
        return pd.read_sql(query, engine, coerce_float=False)


class SQLStaticTableInterface(SQLTableInterface):
    query: str
    gender_space_query_template: str
    race_space_query_template: str

    def __init__(self, query_template: str | None, gender_space_query_template: str, race_space_query_template: str):
        super().__init__(query_template=query_template)
        self.gender_space_query_template = gender_space_query_template
        self.race_space_query_template = race_space_query_template

    @property
    def gender_space_query(self) -> str:
        return open(resources_path(self.gender_space_query_template)).read()

    @property
    def race_space_query(self) -> str:
        return open(resources_path(self.race_space_query_template)).read()

    def load_gender_space_table(self, engine: sqlalchemy.Engine):
        query = self.gender_space_query.format(**COLUMN.as_dict())
        return pd.read_sql(query, engine)

    def load_ethnicity_space_table(self, engine: sqlalchemy.Engine):
        query = self.race_space_query.format(**COLUMN.as_dict())
        return pd.read_sql(query, engine)


class SQLTableResource(TableResource):
    sql_interface: SQLTableInterface

    def __init__(self, columns: TableColumns, query_template: str | None = None):
        super().__init__(columns)
        self.sql_interface = SQLTableInterface(query_template=query_template)

    def load_standard_columns_table(self, engine: sqlalchemy.Engine, *args, **kwargs) -> pd.DataFrame:
        return self.sql_interface.load_standard_columns_table(engine)


class SQLCodedTableResource(CodedTableResource):
    sql_interface: SQLCodedTableInterface

    def __init__(
        self, columns: CodedColumns, query_template: str | None = None, space_query_template: str | None = None
    ):
        super().__init__(columns)
        self.sql_interface = SQLCodedTableInterface(
            query_template=query_template, space_query_template=space_query_template
        )

    def load_standard_columns_table(self, engine: sqlalchemy.Engine, *args, **kwargs) -> pd.DataFrame:
        return self.sql_interface.load_standard_columns_table(engine)

    def load_space_table(self, engine: sqlalchemy.Engine) -> pd.DataFrame:
        return self.sql_interface.load_space_table(engine)


class SQLStaticTableResource(StaticTableResource_MIMICIV):
    columns: StaticTableColumns
    sql_interface: SQLStaticTableInterface

    def __init__(self, query_template: str | None, gender_space_query_template: str, race_space_query_template: str):
        super().__init__()
        self.sql_interface = SQLStaticTableInterface(
            query_template=query_template,
            gender_space_query_template=gender_space_query_template,
            race_space_query_template=race_space_query_template,
        )

    def load_standard_columns_table(self, engine: sqlalchemy.Engine, *args, **kwargs) -> pd.DataFrame:
        return self.sql_interface.load_standard_columns_table(engine)

    def load_gender_space_table(self, engine: sqlalchemy.Engine):
        return self.sql_interface.load_gender_space_table(engine)

    def load_ethnicity_space_table(self, engine: sqlalchemy.Engine):
        return self.sql_interface.load_ethnicity_space_table(engine)


class SQLMixedICDTableResource(MixedICDTableResource_MIMICIV):
    sql_interface: SQLCodedTableInterface

    def __init__(self, query_template: str, space_query_template: str):
        super().__init__()
        self.sql_interface = SQLCodedTableInterface(
            query_template=query_template, space_query_template=space_query_template
        )

    def load_standard_columns_table(self, engine: sqlalchemy.Engine, *args, **kwargs) -> pd.DataFrame:
        return self.sql_interface.load_standard_columns_table(engine)

    def load_space_table(self, engine: sqlalchemy.Engine) -> pd.DataFrame:
        return self.sql_interface.load_space_table(engine)


class SQLMultivariateTimeSeriesResource(MultivariateTimeSeriesTableResource):
    sql_interface: SQLCodedTableInterface

    def __init__(
        self,
        name: str,
        attributes: tuple[str, ...],
        type_hint: tuple[NumericalTypeHint, ...] | None = None,
        default_type_hint: NumericalTypeHint = "N",
        query_template: str | None = None,
    ):
        super().__init__(
            MultivariateTimeSeriesTableMeta(
                name=name, attributes=attributes, type_hint=type_hint, default_type_hint=default_type_hint
            )
        )
        self.sql_interface = SQLCodedTableInterface(
            query_template=query_template, substitutes=dict(attributes=", ".join(attributes))
        )

    def load_standard_columns_table(self, engine: sqlalchemy.Engine, *args, **kwargs) -> pd.DataFrame:
        return self.sql_interface.load_standard_columns_table(engine)


class SQLGroupedMultivariateTimeSeriesTableResource(GroupedMultivariateTimeSeriesTableResource):
    groups: tuple[SQLMultivariateTimeSeriesResource, ...]


ADMISSIONS_CONF = SQLTableResource(query_template="mimiciv/sql/admissions.tsql", columns=AdmissionsTableColumns())
STATIC_CONF = SQLStaticTableResource(
    query_template="mimiciv/sql/static.tsql",
    gender_space_query_template="mimiciv/sql/static_gender_space.tsql",
    race_space_query_template="mimiciv/sql/static_race_space.tsql",
)
DX_DISCHARGE_CONF = SQLMixedICDTableResource(
    query_template="mimiciv/sql/dx_discharge.tsql", space_query_template="mimiciv/sql/dx_discharge_space.tsql"
)

RENAL_OUT_CONF = SQLMultivariateTimeSeriesResource(
    name="renal_out", attributes=("uo_rt_6hr", "uo_rt_12hr", "uo_rt_24hr"), query_template="mimiciv/sql/renal_out.tsql"
)

RENAL_CREAT_CONF = SQLMultivariateTimeSeriesResource(
    name="renal_creat", attributes=("creat",), query_template="mimiciv/sql/renal_creat.tsql"
)

RENAL_AKI_CONF = SQLMultivariateTimeSeriesResource(
    name="renal_aki",
    attributes=("aki_stage_smoothed", "aki_binary"),
    query_template="mimiciv/sql/renal_aki.tsql",
    type_hint=(
        "O",
        "B",
    ),
)  # Ordinal, Binary.

SOFA_CONF = SQLMultivariateTimeSeriesResource(
    name="sofa", attributes=("sofa_24hours",), query_template="mimiciv/sql/sofa.tsql", default_type_hint="O"
)  # Ordinal.
BLOOD_GAS_ATTRIBUTES = (
    "so2",
    "po2",
    "pco2",
    "fio2",
    "fio2_chartevents",
    "aado2",
    "aado2_calc",
    "pao2fio2ratio",
    "ph",
    "baseexcess",
    "bicarbonate",
    "totalco2",
    "hematocrit",
    "hemoglobin",
    "carboxyhemoglobin",
    "methemoglobin",
    "chloride",
    "calcium",
    "temperature",
    "potassium",
    "sodium",
    "lactate",
    "glucose",
)
BLOOD_GAS_CONF = SQLMultivariateTimeSeriesResource(
    name="blood_gas", attributes=BLOOD_GAS_ATTRIBUTES, query_template="mimiciv/sql/blood_gas.tsql"
)

BLOOD_CHEMISTRY_CONF = SQLMultivariateTimeSeriesResource(
    name="blood_chemistry",
    attributes=(
        "albumin",
        "globulin",
        "total_protein",
        "aniongap",
        "bicarbonate",
        "bun",
        "calcium",
        "chloride",
        "creatinine",
        "glucose",
        "sodium",
        "potassium",
    ),
    query_template="mimiciv/sql/blood_chemistry.tsql",
)

CARDIAC_MARKER_CONF = SQLMultivariateTimeSeriesResource(
    name="cardiac_marker",
    attributes=("troponin_t2", "ntprobnp", "ck_mb"),
    query_template="mimiciv/sql/cardiac_marker.tsql",
)

WEIGHT_CONF = SQLMultivariateTimeSeriesResource(
    name="weight", attributes=("weight",), query_template="mimiciv/sql/weight.tsql"
)

CBC_CONF = SQLMultivariateTimeSeriesResource(
    name="cbc",
    attributes=("hematocrit", "hemoglobin", "mch", "mchc", "mcv", "platelet", "rbc", "rdw", "wbc"),
    query_template="mimiciv/sql/cbc.tsql",
)

VITAL_CONF = SQLMultivariateTimeSeriesResource(
    name="vital",
    attributes=(
        "heart_rate",
        "sbp",
        "dbp",
        "mbp",
        "sbp_ni",
        "dbp_ni",
        "mbp_ni",
        "resp_rate",
        "temperature",
        "spo2",
        "glucose",
    ),
    query_template="mimiciv/sql/vital.tsql",
)

# Glasgow Coma Scale, a measure of neurological function
GCS_CONF = SQLMultivariateTimeSeriesResource(
    name="gcs",
    attributes=("gcs", "gcs_motor", "gcs_verbal", "gcs_eyes", "gcs_unable"),
    query_template="mimiciv/sql/gcs.tsql",
    default_type_hint="O",
)  # Ordinal.

# Intracranial pressure
ICP_CONF = SQLMultivariateTimeSeriesResource(name="icp", attributes=("icp",), query_template="mimiciv/sql/icp.tsql")

# Inflammation
INFLAMMATION_CONF = SQLMultivariateTimeSeriesResource(
    name="inflammation", attributes=("crp",), query_template="mimiciv/sql/inflammation.tsql"
)

# Coagulation
COAGULATION_CONF = SQLMultivariateTimeSeriesResource(
    name="coagulation",
    attributes=("pt", "ptt", "inr", "d_dimer", "fibrinogen", "thrombin"),
    query_template="mimiciv/sql/coagulation.tsql",
)
# Blood differential
BLOOD_DIFF_ATTRIBUTES = (
    "neutrophils",
    "lymphocytes",
    "monocytes",
    "eosinophils",
    "basophils",
    "atypical_lymphocytes",
    "bands",
    "immature_granulocytes",
    "metamyelocytes",
    "nrbc",
    "basophils_abs",
    "eosinophils_abs",
    "lymphocytes_abs",
    "monocytes_abs",
    "neutrophils_abs",
)
BLOOD_DIFF_CONF = SQLMultivariateTimeSeriesResource(
    name="blood_diff", attributes=BLOOD_DIFF_ATTRIBUTES, query_template="mimiciv/sql/blood_diff.tsql"
)
ENZYMES_ATTRIBUTES = (
    "ast",
    "alt",
    "alp",
    "ld_ldh",
    "ck_cpk",
    "ck_mb",
    "amylase",
    "ggt",
    "bilirubin_direct",
    "bilirubin_total",
    "bilirubin_indirect",
)
# Enzymes
ENZYMES_CONF = SQLMultivariateTimeSeriesResource(
    name="enzymes", attributes=ENZYMES_ATTRIBUTES, query_template="mimiciv/sql/enzymes.tsql"
)

OBS_COMPONENTS = (
    RENAL_OUT_CONF,
    RENAL_CREAT_CONF,
    RENAL_AKI_CONF,
    SOFA_CONF,
    BLOOD_GAS_CONF,
    BLOOD_CHEMISTRY_CONF,
    CARDIAC_MARKER_CONF,
    WEIGHT_CONF,
    CBC_CONF,
    VITAL_CONF,
    # GCS_CONF,
    ICP_CONF,
    INFLAMMATION_CONF,
    COAGULATION_CONF,
    BLOOD_DIFF_CONF,
    ENZYMES_CONF,
)
OBS_TABLE_CONFIG = SQLGroupedMultivariateTimeSeriesTableResource(groups=OBS_COMPONENTS)

ICU_INPUTS_CONF = SQLCodedTableResource(
    query_template="mimiciv/sql/icu_inputs.tsql",
    space_query_template="mimiciv/sql/icu_inputs_space.tsql",
    columns=AdmissionIntervalRatesTableColumns(),
)
ICU_PROCEDURES_CONF = SQLCodedTableResource(
    query_template="mimiciv/sql/icu_procedures.tsql",
    space_query_template="mimiciv/sql/icu_procedures_space.tsql",
    columns=AdmissionIntervalEventsTableColumns(),
)
HOSP_PROCEDURES_CONF = SQLMixedICDTableResource(
    query_template="mimiciv/sql/hosp_procedures.tsql", space_query_template="mimiciv/sql/hosp_procedures_space.tsql"
)

ENV_MIMICIV_HOST: Final[str] = "MIMICIV_HOST"
ENV_MIMICIV_PORT: Final[str] = "MIMICIV_PORT"
ENV_MIMICIV_USER: Final[str] = "MIMICIV_USER"
ENV_MIMICIV_PASSWORD: Final[str] = "MIMICIV_PASSWORD"
ENV_MIMICIV_DBNAME: Final[str] = "MIMICIV_DBNAME"
ENV_MIMICIV_URL: Final[str] = "MIMICIV_URL"


class SQLMIMICTablesResources(DatasetTablesResources):
    static: SQLStaticTableResource
    admissions: SQLTableResource
    dx_discharge: SQLMixedICDTableResource
    obs: SQLGroupedMultivariateTimeSeriesTableResource
    hosp_procedures: SQLMixedICDTableResource
    icu_procedures: SQLCodedTableResource
    icu_inputs: SQLCodedTableResource

    def __init__(
        self,
        static: SQLStaticTableResource = STATIC_CONF,
        admissions: SQLTableResource = ADMISSIONS_CONF,
        dx_discharge: SQLMixedICDTableResource = DX_DISCHARGE_CONF,
        obs: SQLGroupedMultivariateTimeSeriesTableResource = OBS_TABLE_CONFIG,
        icu_procedures: SQLCodedTableResource = ICU_PROCEDURES_CONF,
        icu_inputs: SQLCodedTableResource = ICU_INPUTS_CONF,
        hosp_procedures: SQLMixedICDTableResource = HOSP_PROCEDURES_CONF,
    ):
        super().__init__(
            static=static,
            admissions=admissions,
            dx_discharge=dx_discharge,
            obs=obs,
            icu_procedures=icu_procedures,
            icu_inputs=icu_inputs,
            hosp_procedures=hosp_procedures,
        )

    @classmethod
    def url(cls) -> str:
        if ENV_MIMICIV_URL in os.environ:
            return os.environ[ENV_MIMICIV_URL]
        elif all(
            e in os.environ
            for e in [ENV_MIMICIV_USER, ENV_MIMICIV_PASSWORD, ENV_MIMICIV_HOST, ENV_MIMICIV_PORT, ENV_MIMICIV_DBNAME]
        ):
            return cls.url_from_credentials(
                user=os.environ[ENV_MIMICIV_USER],
                password=os.environ[ENV_MIMICIV_PASSWORD],
                host=os.environ[ENV_MIMICIV_HOST],
                port=os.environ[ENV_MIMICIV_PORT],
                dbname=os.environ[ENV_MIMICIV_DBNAME],
            )
        else:
            credentials_env_list = [
                ENV_MIMICIV_USER,
                ENV_MIMICIV_PASSWORD,
                ENV_MIMICIV_HOST,
                ENV_MIMICIV_PORT,
                ENV_MIMICIV_DBNAME,
            ]
            raise ValueError(
                f"Environment variables ({ENV_MIMICIV_URL}) or ({', '.join(credentials_env_list)}) are not set."
            )

    @staticmethod
    def url_from_credentials(user: str, password: str, host: str, port: str, dbname: str) -> str:
        return f"postgresql+psycopg://{user}:{password}@{host}:{port}/{dbname}"




import dataclasses
import enum
import json
import logging
import re
from abc import abstractmethod
from collections.abc import Callable, Mapping, Sized
from pathlib import Path
from types import MappingProxyType, NoneType
from typing import Any, cast, Self, TYPE_CHECKING, TypeVar

import equinox as eqx
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
import pandas as pd
import tables as tb

from ._literals import CompressionLibLiteral
from .utils import (
    ArrayTypes,
    equal_arrays,
    load_config,
    NumpyEncoder,
    path_from_getter,
    path_from_jax_keypath,
    write_config,
)


_factory_registry: dict[str, type[eqx.Module]] = {}


class _ModuleMeta(type(eqx.Module)):
    """Meta-class that auto-registers subclasses for factory-based loading.

    Each subclass gains ``__class_key__`` and ``__get_factory__`` for
    round-trippable, type-aware deserialization.
    """
    # This method is called whenever you definite a module: `class Foo(eqx.Module): ...`
    def __new__(
        mcs,
        name,
        bases,
        dict_,
        /,
        strict: bool = False,
        **kwargs,
    ):
        cls = super().__new__(mcs, name, bases, dict_, strict=strict, **kwargs)

        # We need to collect all constructors of the subclasses in one dictionary
        # to use them in deserialization of objects that, themselves, composed of
        # other native types or objects that has the same MetaClass.

        def __class_key__(cls):
            return f"{cls.__module__}.{cls.__qualname__}"

        def __get_factory__(type_str: str):
            try:
                return _factory_registry[type_str]
            except KeyError:
                raise KeyError(
                    f"{type_str} is not a registered type. "
                    f"You may need to import the module that was used during the construction of "
                    f"{type_str} object(s). Registered types ({len(_factory_registry)}): "
                    f"...{list(_factory_registry.keys())[-10:]}. "
                )

        cls.__class_key__ = classmethod(__class_key__)
        cls.__get_factory__ = __get_factory__
        _factory_registry[cls.__class_key__()] = cls  # type: ignore

        return cls


if TYPE_CHECKING:

    class AbstractModule(eqx.Module, metaclass=_ModuleMeta):
        @classmethod
        def __class_key__(cls) -> str:
            return "I am just a placeholder for type checking purposes.."

        @classmethod
        def __get_factory__(cls, type_str: str) -> type[Self]:
            return cls
else:

    class AbstractModule(eqx.Module, metaclass=_ModuleMeta):
        """Base class for all registered modules.

        Subclasses are automatically discoverable during deserialization
        through the registry maintained by ``_ModuleMeta``.
        """


class AbstractHDFSerializable(AbstractModule):
    """Interface for objects that can be serialized into an HDF5 group."""
    @abstractmethod
    def to_hdf_group(self, group: tb.Group) -> None:
        """Write this object into the given HDF5 group."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_hdf_group(cls, group: tb.Group, defer: tuple[tuple[str, ...], ...], levels: int | None) -> Self:
        """Create an instance from an HDF5 group.

        The ``defer`` argument specifies relative paths (as tuples of lambda getter functions) that should
        not be materialized immediately and may be represented by virtual nodes.
        """
        raise NotImplementedError

    @abstractmethod
    def equals(self, other: Self) -> bool:
        """Return True if both objects are equal under the class-specific rules."""
        raise NotImplementedError


class HDFVirtualNode(AbstractHDFSerializable):
    """Placeholder for a value that has not yet been read from HDF5.

    Instances point to the on-disk location and type of a child attribute but
    do not hold the actual data until fetched using ``fetch_at`` or
    ``fetch_all``. Any attempt to access undefined attributes raises with a
    guidance message to fetch first.
    """

    filename: str
    parent_path: str
    type_enum: str
    key: str

    def __init__(self, filename: str, parent_path: str, type_enum: str, key: str = "") -> None:
        self.filename = filename
        self.parent_path = parent_path
        self.type_enum = type_enum
        self.key = key

    def __check_init__(self):
        """Validate types of dataclass fields based on annotations."""
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            assert isinstance(value, cast(type, field.type))

    @property
    def _v_parent_path_seq(self) -> list[str]:  # to a series of directories with the root directory represented by ''
        """Return the split HDF5 path of the parent group as a list of node names."""
        if len(self.parent_path) == 0:
            return []
        if len(self.parent_path) == 1:
            return [""]
        return self.parent_path.split("/")

    def __getattribute__(self, attr: str) -> NoneType:
        """Disallow attribute access on virtual nodes until fetched."""
        try:
            return object.__getattribute__(self, attr)
        except AttributeError:
            raise AttributeError(
                "You are trying to access an attribute in a lazy-loaded node. Please call `fetch_at(..,..)` or "
                "`fetch_all()` on any of the node ancestors first."
            )

    def to_hdf_group(self, group: tb.Group) -> None:
        """Virtual nodes cannot be serialized; instruct user to fetch first."""
        raise ValueError(
            "You are trying to serialize an unfetched node in a PyTree/AbstractHDFSerializable. "
            "Please call `fetch_at(..,..)` or "
            "`fetch_all()` on any of the node ancestors first."
        )

    @classmethod
    def from_hdf_group(
        cls, group: tb.Group, defer: tuple[tuple[str, ...], ...] = (), levels: int | None = None
    ) -> Self:
        """Virtual nodes cannot be directly deserialized."""
        raise ValueError("You are trying to deserialize a VirtualNode.")

    def equals(self, other: AbstractHDFSerializable) -> bool:
        """Virtual nodes cannot participate in equality until fetched."""
        raise ValueError(
            "You are trying to test equality with a virtual unfetched node. Please call `fetch_at(..,..)` or "
            "`fetch_all()` on any of the node ancestors first."
        )


class AbstractConfig(AbstractHDFSerializable):
    """Config objects with JSON-friendly dict views and HDF5 support."""
    @classmethod
    def _map_hierarchical_config(
        cls, unit_config_map: Callable[[Self], dict[str, Any]], x: Any, levels: int | None = None
    ) -> Any:
        if levels is not None and levels <= 0:
            return x
        next_level = levels if levels is None else levels - 1
        if isinstance(x, cls):
            x = unit_config_map(x)
        if isinstance(x, dict):
            return {k: cls._map_hierarchical_config(unit_config_map, v, next_level) for k, v in x.items()}
        elif isinstance(x, list):
            return [cls._map_hierarchical_config(unit_config_map, v, next_level) for v in x]
        elif isinstance(x, tuple):
            return tuple(cls._map_hierarchical_config(unit_config_map, v, next_level) for v in x)
        else:
            return x

    @classmethod
    def _as_normal_dict(cls, x: Self) -> dict[str, Any]:
        return {field.name: getattr(x, field.name) for field in dataclasses.fields(x) if not field.name.startswith("_")}

    @staticmethod
    def _as_typed_dict(x: "AbstractConfig") -> dict[str, Any]:
        return AbstractConfig._as_normal_dict(x) | {"_type": x.__class_key__()}

    @staticmethod
    def _is_typed_dict(x) -> bool:
        return isinstance(x, dict) and "_type" in x

    @staticmethod
    def _map_config_to_dict(unit_map: Callable[[Any], dict[str, Any]], x, levels: int | None = None) -> dict[str, Any]:
        dictionary = AbstractConfig._map_hierarchical_config(unit_map, x, levels)
        if levels is None:
            return json.loads(json.dumps(dictionary, cls=NumpyEncoder))
        return dictionary

    def as_dict(self) -> dict[str, Any]:
        """Return a JSON-serializable dict of fields without type information."""
        return AbstractConfig._map_config_to_dict(AbstractConfig._as_normal_dict, self)

    def as_one_level_dict(self) -> dict[str, Any]:
        """Return a shallow one-level dict view of this config."""
        return AbstractConfig._map_config_to_dict(AbstractConfig._as_normal_dict, self, levels=1)

    def to_dict(self) -> dict[str, Any]:
        """Return a dict that includes ``_type`` for round-trip deserialization."""
        return AbstractConfig._map_config_to_dict(AbstractConfig._as_typed_dict, self)

    def equals(self, other: AbstractHDFSerializable) -> bool:
        assert isinstance(other, AbstractConfig), "Can only compare equality with another AbstractConfig."
        return self.to_dict() == other.to_dict()

    def to_hdf_group(self, group: tb.Group) -> None:
        """Store the typed JSON representation as a raw byte array in HDF5."""
        data = json.dumps(self.to_dict(), cls=NumpyEncoder).encode("utf-8")
        group._v_file.create_array(group, "data", obj=data)

    @classmethod
    def from_hdf_group(
        cls, group: tb.Group, defer: tuple[tuple[str, ...], ...] = (), levels: int | None = None
    ) -> Self:
        assert len(defer) == 0, "Unexpected."
        return cls.from_dict(json.loads(group["data"].read().decode("utf-8")))  # pyright: ignore[reportAttributeAccessIssue] #

    def log_json(self, path: str | Path, key: str):
        """Append this config under ``key`` to a JSON file next to the HDF5 file."""
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        json_path = path.with_suffix(".json")  # config goes here.
        config = load_config(str(json_path)) if json_path.exists() else {}
        config[key] = self.as_dict()  # logging does not aim to produce deserializable config json.
        write_config(config, str(json_path))

    def update(self, other: Self | dict[str, Any]) -> Self:
        """Return a new config by overlaying ``other`` on top of this one.

        !!! Example
        TODO: add example.
        
        """
        if isinstance(other, AbstractConfig):
            other = other.to_dict()
            other["_type"] = self.__class_key__()
        return self.from_dict(self.to_dict() | other)

    @classmethod
    def from_dict(cls, config: dict[str, Any]) -> Self:
        """Construct a config from a nested dict containing ``_type`` markers."""
        def _map_dict_to_config(x):
            if cls._is_typed_dict(x):
                config_class = cls.__get_factory__(x.pop("_type"))
                config_kwargs = {k: _map_dict_to_config(v) for k, v in x.items()}
                return config_class(**config_kwargs)
            elif isinstance(x, dict):
                return {k: _map_dict_to_config(v) for k, v in x.items()}
            elif isinstance(x, list):
                return [_map_dict_to_config(v) for v in x]
            elif isinstance(x, tuple):
                return tuple(_map_dict_to_config(v) for v in x)
            else:
                return x

        return _map_dict_to_config(config)  # pyright: ignore[reportReturnType]

    def path_update(self, path: str, value: Any) -> Self:
        """Create a new config with the attribute at ``path`` replaced by ``value``.

        ``path`` is a dot-separated attribute path; ``value=None`` clears the value.

        !!! Example
        TODO: add example.
        """
        nesting = path.split(".")

        def _get(x):
            for n in nesting:
                x = getattr(x, n)
            return x

        if value is not None:
            return eqx.tree_at(_get, self, type(_get(self))(value))  # pyright: ignore[reportCallIssue]
        return eqx.tree_at(_get, self, None)


class AbstractWithPandasEquivalent(AbstractHDFSerializable):
    """Mixin for objects that can be converted to/from pandas structures."""
    @staticmethod
    def empty_pandas_meta(df: pd.DataFrame | pd.Series) -> pd.DataFrame:
        """Return a one-row metadata table describing an empty DataFrame/Series."""
        meta = {"index_dtype": str(df.index.dtype), "index_name": str(df.index.name)}
        if isinstance(df, pd.Series):
            meta = meta | {"dtype": str(df.dtype), "name": df.name}
        else:
            meta = meta | {f"column_{i}": col for i, col in enumerate(df.columns)}
            meta = meta | {f"column_dtype_{df.columns[i]}": str(dtype) for i, dtype in enumerate(df.dtypes)}
        return pd.DataFrame(meta, index=pd.Index([0]))

    @staticmethod
    def empty_pandas_from_metadata(meta_table: pd.DataFrame) -> pd.DataFrame | pd.Series:
        """Reconstruct an empty DataFrame/Series, preserving all types, columns names, and index name from a metadata table."""
        meta = meta_table.iloc[0].to_dict()
        index_name = meta.pop("index_name")
        index = pd.Index([], dtype=meta.pop("index_dtype"), name=None if index_name == "None" else index_name)
        if "dtype" in meta:
            dtype = meta.pop("dtype")
            name = meta.pop("name")
            name = name if name == name else None  # when None serialized it ended up as a nan
            return pd.Series(dtype=dtype, name=name, index=index)

        cols = []
        column_types = {}
        while len(meta) > 0:
            k, v = meta.popitem()
            if k.startswith("column_dtype_"):
                column_types[k.split("column_dtype_")[1]] = v
            elif k.startswith("column_"):
                order = int(k.split("column_")[1])
                cols.append((order, v))
        cols = pd.Series([col for _, col in sorted(cols, key=lambda x: x[0])])
        return pd.DataFrame(columns=cols, index=index).astype(column_types)

    @classmethod
    def serialize_pandas(cls, data: pd.DataFrame | pd.Series, group: tb.Group):
        """Persist pandas data to HDF5, preserving empty-shape metadata when needed."""
        # Empty dataframes/series are not saved in hdf file, https://github.com/pandas-dev/pandas/issues/13016,
        # https://github.com/PyTables/PyTables/issues/592
        # We still need to preserve the column names, dtypes, index name, index dtype, in a metadata object, to
        # fulfill perfect serialisation/deserialisation and have stricter unit-testing.
        hdf = group._v_file
        if data.empty:
            key = hdf.create_group(group, "empty_metadata")._v_pathname
            cls.empty_pandas_meta(data).to_hdf(hdf.filename, key=key, format="table")
        else:
            data.to_hdf(hdf.filename, key=group._v_pathname, format="table")

    @classmethod
    def deserialize_pandas(cls, store: tb.Group) -> pd.DataFrame | pd.Series:
        """Load pandas data from HDF5, handling the empty-data case via metadata."""
        hdf = store._v_file
        if "empty_metadata" in store:
            meta_table = pd.read_hdf(hdf.filename, key=store.empty_metadata._v_pathname)
            assert isinstance(meta_table, pd.DataFrame)
            return cls.empty_pandas_from_metadata(meta_table)
        else:
            table_or_series = pd.read_hdf(hdf.filename, key=store._v_pathname)
            assert isinstance(table_or_series, (pd.DataFrame, pd.Series))
            return table_or_series

    @abstractmethod
    def to_pandas(self) -> pd.DataFrame | pd.Series:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_pandas(cls, pandas: pd.DataFrame | pd.Series) -> Self:
        raise NotImplementedError

    def to_hdf_group(self, group: tb.Group) -> None:
        """Store the pandas-equivalent representation under a ``data`` subgroup."""
        group._v_attrs.classname = self.__class_key__()
        hdf_file = group._v_file
        self.serialize_pandas(self.to_pandas(), hdf_file.create_group(group, "data"))

    @classmethod
    def from_hdf_group(
        cls, group: tb.Group, defer: tuple[tuple[str, ...], ...] = (), levels: int | None = None
    ) -> Self:
        assert len(defer) == 0, "Unexpected."
        classname = group._v_attrs.classname.item()
        hdf_file = group._v_file
        return cls.__get_factory__(classname).from_pandas(cls.deserialize_pandas(hdf_file.get_node(group, "data")))

    def equals(self, other: AbstractHDFSerializable) -> bool:
        """Equality based on pandas ``.equals`` for the underlying data."""
        assert isinstance(other, AbstractWithPandasEquivalent), "Can only compare with a compatible pandas-equivalent."
        return type(self) is type(other) and self.to_pandas().equals(other.to_pandas())


class AbstractWithDataframeEquivalent(AbstractWithPandasEquivalent):
    """Mixin specializing the pandas equivalent to a DataFrame."""
    @abstractmethod
    def to_dataframe(self) -> pd.DataFrame:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_dataframe(cls, dataframe: pd.DataFrame) -> Self:
        raise NotImplementedError

    def to_pandas(self) -> pd.DataFrame:
        return self.to_dataframe()

    @classmethod
    def from_pandas(cls, pandas: pd.DataFrame | pd.Series) -> Self:
        assert isinstance(pandas, pd.DataFrame), "Expected a pandas DataFrame."
        return cls.from_dataframe(pandas)


class AbstractWithSeriesEquivalent(AbstractWithPandasEquivalent):
    """Mixin specializing the pandas equivalent to a Series."""
    @abstractmethod
    def to_series(self) -> pd.Series:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_series(cls, dataframe: pd.Series) -> Self:
        raise NotImplementedError

    def to_pandas(self) -> pd.Series:
        return self.to_series()

    @classmethod
    def from_pandas(cls, pandas: pd.DataFrame | pd.Series) -> Self:
        assert isinstance(pandas, pd.Series), "Expected a pandas Series."
        return cls.from_series(pandas)


class SERIALIZABLE_FIELD(enum.Enum):
    """Enumeration of supported field categories for HDF5 serialization."""
    none = (type(None),)
    numpy_array = tuple(ArrayTypes)
    pandas_dataframe = (pd.DataFrame,)
    pandas_series = (pd.Series,)
    pandas_equivalent = (AbstractWithPandasEquivalent,)
    hdf_serializable = (AbstractHDFSerializable,)
    config = (AbstractConfig,)
    string = (str,)
    integer = (int,)
    float = (float,)
    boolean = (bool,)
    timestamp = (pd.Timestamp,)
    list = (list,)
    tuple = (tuple,)
    dict = (dict,)
    set = (set,)
    frozenset = (frozenset,)
    mapping_proxy = (MappingProxyType,)


COMPARISON_PRIORITY = MappingProxyType(
    {e.name: 100 for e in SERIALIZABLE_FIELD}
    | {  # default: 100
        e.name: 0
        for e in (SERIALIZABLE_FIELD.none,)  # begin with None
    }
    | {
        e.name: 1
        for e in (
            SERIALIZABLE_FIELD.boolean,
            SERIALIZABLE_FIELD.integer,
            SERIALIZABLE_FIELD.float,
            SERIALIZABLE_FIELD.timestamp,
        )  # then scalers
    }
    | {
        e.name: 2
        for e in (SERIALIZABLE_FIELD.string,)  # then variable-length strings
    }
    | {
        e.name: 3
        for e in (SERIALIZABLE_FIELD.config,)  # then config
    }
    | {
        e.name: 4
        for e in (  # then intensive data.
            SERIALIZABLE_FIELD.numpy_array,
            SERIALIZABLE_FIELD.pandas_dataframe,
            SERIALIZABLE_FIELD.pandas_series,
            SERIALIZABLE_FIELD.pandas_equivalent,
        )
    }
)  # collections, mappings, and HDFSerializable subclasses are left because they can be nested with
# other HDFSerializables
assert all(isinstance(e.value, tuple) and all(isinstance(t, type) for t in e.value) for e in SERIALIZABLE_FIELD), (
    "Expected a tuple of types."
)

_TYPE_ENUM_DICT: MappingProxyType[type, str] = MappingProxyType(
    {t: e.name for e in SERIALIZABLE_FIELD for t in e.value}
)
SERIALIZABLE_FIELD_TYPES = tuple(_TYPE_ENUM_DICT.keys())
SERIALIZABLE_FLAT_COLLECTION = (
    SERIALIZABLE_FIELD.set,
    SERIALIZABLE_FIELD.list,
    SERIALIZABLE_FIELD.tuple,
    SERIALIZABLE_FIELD.frozenset,
)
SERIALIZABLE_FLAT_DICT = (SERIALIZABLE_FIELD.dict, SERIALIZABLE_FIELD.mapping_proxy)
SERIALIZABLE_FLAT_DICT_KEY = (SERIALIZABLE_FIELD.integer, SERIALIZABLE_FIELD.string)

SERIALIZABLE_FLAT_COLLECTION_TYPES: tuple[type, ...] = sum((e.value for e in SERIALIZABLE_FLAT_COLLECTION), ())
SERIALIZABLE_FLAT_DICT_TYPES: tuple[type, ...] = sum((e.value for e in SERIALIZABLE_FLAT_DICT), ())
SERIALIZABLE_FLAT_DICT_KEY_TYPES: tuple[type, ...] = sum((e.value for e in SERIALIZABLE_FLAT_DICT_KEY), ())

# Serialised elements within homogeneous- and flat-list, -tuple, -dict.
SERIALIZABLE_ELEMENT = (
    SERIALIZABLE_FIELD.numpy_array,
    SERIALIZABLE_FIELD.pandas_dataframe,
    SERIALIZABLE_FIELD.pandas_series,
    SERIALIZABLE_FIELD.pandas_equivalent,
    SERIALIZABLE_FIELD.hdf_serializable,
    SERIALIZABLE_FIELD.config,
    SERIALIZABLE_FIELD.string,
    SERIALIZABLE_FIELD.integer,
    SERIALIZABLE_FIELD.float,
    SERIALIZABLE_FIELD.boolean,
    SERIALIZABLE_FIELD.timestamp,
)
SERIALIZABLE_ELEMENT_TYPES = sum((e.value for e in SERIALIZABLE_ELEMENT), ())

# These element types contained within a homogeneous container, can be converted to a pd.Series first.
SERIES_GROUPED_ELEMENT = (
    SERIALIZABLE_FIELD.float,
    SERIALIZABLE_FIELD.integer,
    SERIALIZABLE_FIELD.boolean,
    SERIALIZABLE_FIELD.string,
    SERIALIZABLE_FIELD.none,
)
SERIES_GROUPED_ELEMENT_TYPES = sum((e.value for e in SERIES_GROUPED_ELEMENT), ())


class _MyCustomList_set(list):
    # a custom list used to deal with replacing sets with a list, since sets are not considered a pytree
    # like lists/dicts/tuples, so we replace all sets with this type of list, so we can reverse the operation
    # (i.e. map back to a set) by identifying this type in the pytree. See `fetch_all` for more details.
    pass


class _MyCustomList_frozenset(list):
    # same but for frozenset.
    pass


jtu.register_pytree_node(_MyCustomList_set, lambda x: (list(x), None), lambda _, x: _MyCustomList_set(x))
jtu.register_pytree_node(_MyCustomList_frozenset, lambda x: (list(x), None), lambda _, x: _MyCustomList_frozenset(x))

MAX_SEGMENT_SIZE = 500


class AbstractVxData(AbstractHDFSerializable):
    """Base class providing field-wise HDF5 serialization and lazy-loading.

    Subclasses must be constructible from their dataclass-like fields so that
    deserialization can materialize instances via ``__init__(**fields)``.
    """

    def __check_init__(self):
        for f in (k for k in self.fields):
            obj = getattr(self, f)
            if obj is None:
                continue
            try:
                _ = self.object_type_enum_name(obj)
            except (KeyError, StopIteration):
                assert False, (
                    f"Unsupported field type {type(obj)} for attribute {f}. It must be a type in "
                    f"{tuple(t.__qualname__ for t in SERIALIZABLE_FIELD_TYPES)} or a subclass of "
                    f"AbstractHDFSerializable."
                )

    @property
    def fields(self) -> tuple[str, ...]:
        """Return the names of the class fields."""
        return tuple(k.name for k in dataclasses.fields(self))

    @property
    def comparison_prioritized_fields(self) -> tuple[str, ...]:
        """Return field names ordered by comparison priority for equality checks."""
        return tuple(
            sorted(self.fields, key=lambda f: COMPARISON_PRIORITY[self.object_type_enum_name(getattr(self, f))])
        )

    @classmethod
    def object_type_enum_name(cls, obj: Any) -> str:
        """Map a value to its ``SERIALIZABLE_FIELD`` enum name."""
        if type(obj) in _TYPE_ENUM_DICT:
            return _TYPE_ENUM_DICT[type(obj)]
        if type(obj) in (_MyCustomList_set, _MyCustomList_frozenset):
            return _TYPE_ENUM_DICT[type(list())]
        elif isinstance(obj, SERIALIZABLE_FIELD.config.value):
            return SERIALIZABLE_FIELD.config.name
        elif isinstance(obj, SERIALIZABLE_FIELD.pandas_equivalent.value):
            return SERIALIZABLE_FIELD.pandas_equivalent.name
        ## Potentially a subclass of AbstractHDFSerializable
        elif isinstance(obj, SERIALIZABLE_FIELD.hdf_serializable.value):
            return SERIALIZABLE_FIELD.hdf_serializable.name
        ## This is for PipelineReportTable.
        elif isinstance(obj, SERIALIZABLE_FIELD.pandas_dataframe.value):
            return SERIALIZABLE_FIELD.pandas_dataframe.name
        ## This is for jax Array subclasses (e.g. jax.ArrayImpl)
        elif isinstance(obj, SERIALIZABLE_FIELD.numpy_array.value):
            return SERIALIZABLE_FIELD.numpy_array.name
        else:
            raise ValueError(f"Unsupported type {type(obj)}.")

    def equals(self, other: AbstractHDFSerializable) -> bool:
        # Need stricter than equinox's `equal_trees(... ,typematch=True)`; For example, ensures pandas.DataFrame
        # objects are compared with `equals` instead of `__eq__`.
        assert isinstance(other, AbstractVxData), "Can only compare equality with another AbstractVxData."
        return type(self) is type(other) and self.fields == other.fields and self.equal_attributes(self, other)

    @staticmethod
    def equal_attributes(self_obj: AbstractHDFSerializable, other_obj: AbstractHDFSerializable) -> bool:
        def _equal_attributes(a: Any, b: Any):
            if type(a) is not type(b):
                return False
            type_enum_name = self_obj.object_type_enum_name(a)
            match SERIALIZABLE_FIELD[type_enum_name]:
                case (
                    SERIALIZABLE_FIELD.boolean
                    | SERIALIZABLE_FIELD.integer
                    | SERIALIZABLE_FIELD.float
                    | SERIALIZABLE_FIELD.timestamp
                    | SERIALIZABLE_FIELD.none
                    | SERIALIZABLE_FIELD.string
                ):
                    if a != b:
                        return False
                case SERIALIZABLE_FIELD.numpy_array:
                    if not equal_arrays(a, b):
                        return False
                case SERIALIZABLE_FIELD.pandas_dataframe | SERIALIZABLE_FIELD.pandas_series:
                    if a.index.name != b.index.name:
                        return False
                    if isinstance(a, pd.Series) and (a.name != b.name):
                        return False
                    if not a.equals(b):
                        return False
                case (
                    SERIALIZABLE_FIELD.config
                    | SERIALIZABLE_FIELD.hdf_serializable
                    | SERIALIZABLE_FIELD.pandas_equivalent
                ):
                    if not a.equals(b):
                        return False

                case (
                    SERIALIZABLE_FIELD.list
                    | SERIALIZABLE_FIELD.tuple
                    | SERIALIZABLE_FIELD.set
                    | SERIALIZABLE_FIELD.frozenset
                ):
                    if len(a) != len(b):
                        return False
                    if isinstance(a, set):  # order is not guaranteed!.
                        a, b = sorted(a, key=hash), sorted(b, key=hash)
                    for a_item, b_item in zip(a, b):
                        if not _equal_attributes(a_item, b_item):
                            return False
                case SERIALIZABLE_FIELD.dict | SERIALIZABLE_FIELD.mapping_proxy:
                    if len(a) != len(b):
                        return False

                    a_keys, a_values = a.keys(), a.values()
                    b_keys, b_values = b.keys(), b.values()
                    if a_keys != b_keys:
                        return False
                    for a_item, b_item in zip(a_values, b_values):
                        if not _equal_attributes(a_item, b_item):
                            return False
                case _:
                    raise ValueError(f"Unhandled type {type_enum_name} for attribute {attribute}.")
            return True

        for attribute in self_obj.comparison_prioritized_fields:
            self_value, other_value = getattr(self_obj, attribute), getattr(other_obj, attribute)
            if not _equal_attributes(self_value, other_value):
                return False

        return True

    @classmethod
    def serialize_object(cls, parent_group: tb.Group, obj: Any, hdf_key: str):
        """Serialize a single object under ``hdf_key`` attached to ``parent_group``."""
        # Don't create a group for native-type attributes (float, int, boolean, string).
        # Attach their values as attributes to the parent group.
        hdf = parent_group._v_file
        group = lambda: hdf.create_group(parent_group, hdf_key)
        match SERIALIZABLE_FIELD[cls.object_type_enum_name(obj)]:
            case SERIALIZABLE_FIELD.numpy_array:
                hdf.create_array(parent_group, hdf_key, obj=obj)
            case SERIALIZABLE_FIELD.pandas_dataframe | SERIALIZABLE_FIELD.pandas_series:
                AbstractWithPandasEquivalent.serialize_pandas(obj, group())
            case SERIALIZABLE_FIELD.hdf_serializable | SERIALIZABLE_FIELD.config | SERIALIZABLE_FIELD.pandas_equivalent:
                obj.to_hdf_group(group())
            case (
                SERIALIZABLE_FIELD.integer
                | SERIALIZABLE_FIELD.float
                | SERIALIZABLE_FIELD.boolean
                | SERIALIZABLE_FIELD.string
                | SERIALIZABLE_FIELD.none
            ):
                parent_group._v_attrs[hdf_key] = obj
            case SERIALIZABLE_FIELD.timestamp:
                hdf.create_array(parent_group, hdf_key, obj=obj.value)
            case (
                SERIALIZABLE_FIELD.list
                | SERIALIZABLE_FIELD.tuple
                | SERIALIZABLE_FIELD.set
                | SERIALIZABLE_FIELD.frozenset
            ):
                cls.serialize_sequence(group(), list(obj))
            case SERIALIZABLE_FIELD.dict | SERIALIZABLE_FIELD.mapping_proxy:
                cls.serialize_dict(group(), obj)
            case _:
                raise ValueError(f"Unknown type {type(obj)} for attribute {hdf_key}")

    @classmethod
    def deserialize_object(
        cls,
        parent_group: tb.Group,
        hdf_key: str,
        attr_type_enum_name: str,
        defer: tuple[tuple[str, ...], ...],
        levels: int | None,
    ):
        """Inverse of ``serialize_object`` supporting deferral and virtual nodes."""
        hd_file = parent_group._v_file
        node = lambda: hd_file.get_node(parent_group, hdf_key)
        defer_current = any(len(d) == 1 and d[0] == hdf_key for d in defer)
        defer_next = tuple(d[1:] for d in defer if len(d) > 1 and d[0] == hdf_key)

        if defer_current or levels == 0:
            return HDFVirtualNode(hd_file.filename, parent_group._v_pathname, attr_type_enum_name, key=hdf_key)

        next_level = levels - 1 if levels is not None else None

        match SERIALIZABLE_FIELD[attr_type_enum_name]:
            case SERIALIZABLE_FIELD.numpy_array:
                return node().read()
            case SERIALIZABLE_FIELD.pandas_dataframe | SERIALIZABLE_FIELD.pandas_series:
                return AbstractWithPandasEquivalent.deserialize_pandas(node())
            case SERIALIZABLE_FIELD.pandas_equivalent:
                return AbstractWithPandasEquivalent.from_hdf_group(node(), defer_next, next_level)
            case SERIALIZABLE_FIELD.config:
                return AbstractConfig.from_hdf_group(node(), defer_next, next_level)
            case SERIALIZABLE_FIELD.hdf_serializable:
                return cls.from_hdf_group(node(), defer_next, next_level)
            case SERIALIZABLE_FIELD.none:
                return None
            case (
                SERIALIZABLE_FIELD.integer
                | SERIALIZABLE_FIELD.float
                | SERIALIZABLE_FIELD.boolean
                | SERIALIZABLE_FIELD.string
            ):
                value = parent_group._v_attrs[hdf_key]
                return value.item() if value is not None else None
            case SERIALIZABLE_FIELD.timestamp:
                return pd.Timestamp(node().read())
            case (
                SERIALIZABLE_FIELD.list
                | SERIALIZABLE_FIELD.tuple
                | SERIALIZABLE_FIELD.set
                | SERIALIZABLE_FIELD.frozenset
            ):
                (sequence_type,) = SERIALIZABLE_FIELD[attr_type_enum_name].value
                assert sequence_type in (list, tuple, set, frozenset)
                seq = cls.deserialize_sequence(node(), defer_next, next_level)
                return sequence_type(seq)
            case SERIALIZABLE_FIELD.dict | SERIALIZABLE_FIELD.mapping_proxy:
                (collection_type,) = SERIALIZABLE_FIELD[attr_type_enum_name].value
                c = cls.deserialize_dict(node(), defer_next, next_level)
                assert collection_type in (dict, MappingProxyType)
                return collection_type(c)
            case _:
                raise ValueError(f"Unhandled type {attr_type_enum_name} for attribute {hdf_key}.")

    @classmethod
    def make_hdf_key(cls, item: str | int):
        """Create a stable HDF5 key for a collection element index/key."""
        return f"key_{item}"

    @classmethod
    def make_hdf_segment_key(cls, item: str | int) -> str:
        """Key used to group large collections into segments for performance."""
        return f"_x_segment_{item}"

    @classmethod
    def create_bookkeeping_segments(
        cls, parent_group: tb.Group, entries: list[str | int], objects: list[Any]
    ) -> tuple[dict[str, tb.Group], pd.DataFrame]:
        """Create per-segment groups and a metadata table for heterogeneous collections."""
        # pytables raises a performance warning when the number of children exceeds 16
        assert len(entries) != 0 and len(entries) == len(objects)
        hdf_keys = list(map(cls.make_hdf_key, entries))
        types = list(map(cls.object_type_enum_name, objects))
        metadata = pd.DataFrame({"hdf_key": hdf_keys, "type": types}, index=pd.Index(entries))
        metadata["segment"] = [cls.make_hdf_segment_key(i // MAX_SEGMENT_SIZE) for i in range(len(entries))]
        if metadata["segment"].nunique() > 1:
            h5file = parent_group._v_file
            segmented_groups = {k: h5file.create_group(parent_group, k) for k in metadata["segment"].unique()}
        else:
            segmented_groups = {cls.make_hdf_segment_key(0): parent_group}
        return segmented_groups, metadata

    @classmethod
    def get_bookkeeping_segments(cls, parent_group: tb.Group, metadata: pd.DataFrame) -> dict[str, tb.Group]:
        """Resolve or create segment groups referenced by the metadata table."""
        if metadata["segment"].nunique() == 1:
            return {cls.make_hdf_segment_key(0): parent_group}
        else:
            hf5_file = parent_group._v_file
            return {k: hf5_file.get_node(parent_group, k, "Group") for k in metadata["segment"].unique()}

    @classmethod
    def serialize_sequence(cls, group: tb.Group, sequence: list[Any]):
        """Serialize a list-like sequence, grouping scalar-like items as a Series when possible."""
        if len(sequence) == 0:
            return
        if set(map(type, sequence)).issubset(SERIES_GROUPED_ELEMENT_TYPES):
            cls.serialize_object(group, pd.Series(sequence), "data")
        else:
            cls.serialize_heterogeneous_collection(group, dict(zip(range(len(sequence)), sequence)))

    @classmethod
    def serialize_dict(cls, group: tb.Group, d: Mapping[str | int, Any]):
        """Serialize a mapping, grouping scalar-like values as a Series when possible."""
        if len(d) == 0:
            return
        if set(map(type, d.values())).issubset(SERIES_GROUPED_ELEMENT_TYPES):
            cls.serialize_object(group, pd.Series(d), "data")
        else:
            cls.serialize_heterogeneous_collection(group, d)

    @classmethod
    def serialize_heterogeneous_collection(cls, group: tb.Group, data: Mapping[str | int, Any]):
        """Serialize a mapping of heterogeneously-typed items with segmentation metadata."""
        segment_group, metadata = cls.create_bookkeeping_segments(group, list(data.keys()), list(data.values()))
        cls.serialize_object(group, metadata, "metadata")
        for segment, hdf_key, key in zip(metadata["segment"], metadata["hdf_key"], metadata.index):
            cls.serialize_object(segment_group[segment], data[key], hdf_key)

    @classmethod
    def deserialize_sequence(cls, group: tb.Group, defer: tuple[tuple[str, ...], ...], levels: int | None) -> list[Any]:
        """Deserialize a sequence previously stored by ``serialize_sequence``."""
        if group._v_nchildren == 0:
            return []
        if "data" in group:
            series = pd.read_hdf(group._v_file.filename, key=group.data._v_pathname)
            assert isinstance(series, pd.Series)
            return series.values.tolist()
        return list(cls.deserialize_heterogeneous_collection(group, defer, levels).values())

    @classmethod
    def deserialize_dict(
        cls, group: tb.Group, defer: tuple[tuple[str, ...], ...], levels: int | None
    ) -> dict[str | int, Any]:
        """Deserialize a mapping previously stored by ``serialize_dict``."""
        if group._v_nchildren == 0:
            return {}
        elif "data" in group:
            series = pd.read_hdf(group._v_file.filename, key=group.data._v_pathname)
            assert isinstance(series, pd.Series)
            return series.to_dict()

        return cls.deserialize_heterogeneous_collection(group, defer, levels)

    @classmethod
    def deserialize_heterogeneous_collection(
        cls, group: tb.Group, defer: tuple[tuple[str, ...], ...], levels: int | None
    ) -> dict[str | int, Any]:
        """Deserialize a heterogeneously-typed mapping using its segmentation metadata."""
        meta = pd.read_hdf(group._v_file.filename, key=group.metadata._v_pathname)
        assert isinstance(meta, pd.DataFrame)
        segment_group = cls.get_bookkeeping_segments(group, meta)
        return {
            k: cls.deserialize_object(segment_group[s], hdf_k, t, defer, levels)
            for k, hdf_k, s, t in zip(meta.index, meta["hdf_key"], meta["segment"], meta["type"])
        }

    def to_hdf_group(self, group: tb.Group) -> None:
        """Serialize all fields into the given group."""
        group._v_attrs.classname = self.__class_key__()
        # Store the types enum for each attribute as a pd.Series (directly equivalent to a dictionary).
        fields = self.fields
        values = [getattr(self, attribute) for attribute in fields]
        group._v_attrs.type_enum = json.dumps(dict(zip(fields, map(self.object_type_enum_name, values))))
        for attribute, obj in zip(fields, values):
            self.serialize_object(group, obj, attribute)

    @classmethod
    def _from_hdf_group(cls, group: tb.Group, defer: tuple[tuple[str, ...], ...], levels: int | None) -> Self:
        """Instantiate this class by reading all fields serialized by ``to_hdf_group``."""
        type_enum = json.loads(group._v_attrs.type_enum)
        data = {
            attr: cls.deserialize_object(group, attr, attr_type_enum, defer, levels)
            for attr, attr_type_enum in type_enum.items()
        }
        return cls(**data)

    @classmethod
    def from_hdf_group(
        cls, group: tb.Group, defer: tuple[tuple[str, ...], ...] = (), levels: int | None = None
    ) -> Self:
        """Factory-based loader that dispatches based on stored classname."""
        classname = group._v_attrs.classname.item()
        return cls.__get_factory__(classname)._from_hdf_group(group, defer, levels=levels)

    def save(
        self,
        store: str | Path | tb.Group,
        complib: CompressionLibLiteral = "blosc",
        complevel: int = 9,
        log_config_json: bool = True,
    ):
        """Persist the object to an HDF5 file or group.

        If ``store`` is a path, the file ``<store>.h5`` is created. When
        ``log_config_json`` is True, any ``AbstractConfig`` fields are also
        written next to the HDF5 as a JSON file for convenience.

        Parameters:
            store: Path to the HDF5 file or a PyTables group. If a path, the file ``<store>.h5`` is created.
            complib: Compression library to use. Default is ``blosc``.
            complevel: Compression level to use. Default is ``9``.
            log_config_json: Whether to log the config as a JSON file. Default is ``True``.

        Returns:
            None


        !!! Example
        TODO: add example.

        !!! Example
        TODO: add another example.

        !!! Example
        TODO: add another example.

        """
        # complib: Literal['blosc', 'zlib', 'lzo', 'bzip2'] = 'blosc', complevel: int = 9
        # lzo lvl1 is a good compromise between speed and compression ratio.
        # https://www.pytables.org/usersguide/optimization.html (Figure 15).
        if not isinstance(store, tb.Group):
            filters = tb.Filters(complib=complib, complevel=complevel)
            with tb.open_file(
                str(Path(store).with_suffix(".h5")),
                mode="w",
                filters=filters,
                max_numexpr_threads=None,
                max_blosc_threads=None,
            ) as store:
                return self.save(store.root, log_config_json=log_config_json)
        if log_config_json:
            config_fields = [c for c in self.fields if isinstance(getattr(self, c), AbstractConfig)]
            filename = store._v_file.filename
            for key, config in zip(config_fields, map(lambda c: getattr(self, c), config_fields)):
                config.log_json(filename, key=key)

        self.to_hdf_group(store)

    @classmethod
    def load(
        cls,
        hf5_filename_or_group: str | Path | tb.Group,
        defer: tuple[Callable[[Self], Any], ...] = (),
        levels: tuple[int, ...] | None = None,
    ) -> Self:
        """Load an instance from HDF5 with optional deferral and subtree materialization limits.

        Parameters:
            hf5_filename_or_group: Path to the HDF5 file or a PyTables group.
            defer: Tuple of getters selecting attributes to remain virtual.
            levels: If given with ``defer``, fetch only this many levels deep.

        Returns:
            An instance of the class.

        !!! Example
        TODO: add example.

        !!! Example
        TODO: add another example.

        !!! Example
        TODO: add another example.
        """
        if not isinstance(hf5_filename_or_group, tb.Group):
            with tb.open_file(str(Path(hf5_filename_or_group).with_suffix(".h5")), "r") as hf5_file:
                return cls.load(hf5_file.root, defer, levels=levels)

        defer_paths = tuple(tuple(path_from_getter(d, getitem_transform=cls.make_hdf_key)) for d in defer)
        loaded = cls.from_hdf_group(hf5_filename_or_group, defer_paths)
        if levels is not None:
            return fetch_at(defer, loaded, levels=levels)
        return loaded

    def to_numpy_arrays(self):
        """Convert all array leaves to NumPy arrays and return a new instance."""
        arrs, others = eqx.partition(self, eqx.is_array)
        arrs = jtu.tree_map(lambda a: np.array(a), arrs)
        return eqx.combine(arrs, others)

    def to_jax_arrays(self):
        """Convert all array leaves to JAX arrays and return a new instance."""
        arrs, others = eqx.partition(self, eqx.is_array)
        arrs = jtu.tree_map(lambda a: jnp.array(a), arrs)
        return eqx.combine(arrs, others)


def _match_child_parent_paths(ch: list[str], pt: list[str]):
    """Return True if child and parent path segments are consistent for fetches."""
    # For example:
    # - if we hold a node representing the _patients_, and we want to fetch
    #   a v_node representing the observables of the third admission of the patient at index 6, and
    # - if the _patients_ itself has a sequence of ancestors
    #   (a.b.....y.z), such that the observables accessed with the expression:
    #       a.b.....y.z.patients[key].admissions[index].observables, then:
    # - child (relative to _patients_) = ["6", "admissions", "2", "observables"]
    # - parent (relative to the child) = ["a", "b", ..., "x", "y", "z", "patients", "6", "admissions", "2"]
    if len(ch) == 1:
        return True
    # Remove any segment node from parent.
    segment_pattern = re.compile(AbstractVxData.make_hdf_segment_key(r"\d+"))
    pt = [pti for pti in pt if re.match(segment_pattern, pti) is None]

    child_overlap = ch[:-1]
    parent_overlap = pt[-(len(ch) - 1) :]
    return child_overlap == parent_overlap


T = TypeVar("T", bound=AbstractVxData)
type HDFVirtualNodeGet[T] = Callable[[T], HDFVirtualNode]


def fetch_at[T](
    where: HDFVirtualNodeGet[T] | tuple[HDFVirtualNodeGet[T], ...],
    tree: T,
    levels: int | None | tuple[int | None, ...] = None,
) -> T:
    """Fetch one or more virtual nodes from disk into ``tree``.

    Parameters:
        where: Getter or tuple of getters selecting the virtual node(s).
        tree: Root object containing the virtual node(s).
        levels: If provided, controls depth of subtree materialization for each
            getter (single int applies to all; tuple applies per getter).

    Returns:
        The root object with the virtual node(s) fetched.

    !!! Example
    TODO: add example.

    !!! Example
    TODO: add another example.

    !!! Example
    TODO: add another example.
    """
    # deal with a sequence to avoid opening a file for each v_node fetch.
    if callable(where):
        where = (where,)
    if isinstance(levels, Sized):
        levels = tuple(levels)
    else:
        return fetch_at(where, tree, (levels,) * len(where))

    if len(where) == 0:
        return tree

    assert len(where) == len(levels), (
        f"Passed a tuple of getters and a tuple of levels of different sizes: {len(where)} and {len(levels)}."
    )

    if any(not isinstance(w(tree), HDFVirtualNode) for w in where):
        logging.warning("You are trying to fetch a non-virtual node, which might be already fetched.")

    v_nodes = [w(tree) for w in where]
    v_nodes_paths = [path_from_getter(w, getitem_transform=AbstractVxData.make_hdf_key) for w in where]
    parent_paths = [v_node._v_parent_path_seq for v_node in v_nodes]

    assert all(map(_match_child_parent_paths, v_nodes_paths, parent_paths)), (
        "Incompatible parent path and v_node path and parent path."
    )
    assert all(v_node.filename == v_nodes[0].filename for v_node in v_nodes), "Unexpected Filename mismatch!"
    with tb.open_file(v_nodes[0].filename, "r") as hf5_file:
        for w, v_node, hdf_key, levels_i in zip(where, v_nodes, [p[-1] for p in v_nodes_paths], levels):
            parent_group = hf5_file.get_node(v_node.parent_path)
            assert isinstance(parent_group, tb.Group)
            fetched = AbstractVxData.deserialize_object(
                parent_group, hdf_key=hdf_key, attr_type_enum_name=v_node.type_enum, defer=(), levels=levels_i
            )
            tree = eqx.tree_at(w, tree, fetched)
        return tree
    raise RuntimeError(f"Failed to open the HDFFile: {v_nodes[0].filename}.")


def fetch_one_level_at[T](where: HDFVirtualNodeGet[T] | tuple[HDFVirtualNodeGet[T], ...], tree: T) -> T:
    """Convenience wrapper to fetch exactly one level at selected locations.
    
    Parameters:
        where: Getter or tuple of getters selecting the virtual node(s).
        tree: Root object containing the virtual node(s).

    Returns:
        The root object with the virtual node(s) fetched.

    !!! Example
    TODO: add example.

    !!! Example
    TODO: add another example.

    !!! Example
    TODO: add another example.
    """
    # Useful to fetch dictionary keys with virtual nodes for values.
    # Or a sequence of vitruals, or object with virtual nodes for attributes.
    return fetch_at(where, tree=tree, levels=1)


def fetch_all[T](tree: T) -> T:
    """Fetch all virtual nodes anywhere in the tree, preserving sets/frozensets.
    
    Parameters:
        tree: Root object containing the virtual node(s).

    Returns:
        The root object with the virtual node(s) fetched.

    !!! Example
    TODO: add example.

    !!! Example
    TODO: add another example.

    !!! Example
    TODO: add another example.
    """
    # Note 1:
    # Preprocessing to catch any set in the pytree. JAX pytree does not
    # navigate into sets as it does with list/dicts/tuples.
    # We will transform any such set into a list, then we return them to sets after fetching.
    # Maybe an alternative is to enforce a design where lazy loading is not allowed with sets involved.

    is_set = jtu.tree_map(lambda a: isinstance(a, set), tree)
    is_frozenset = jtu.tree_map(lambda a: isinstance(a, frozenset), tree)
    apply = (
        lambda a, isset, isfrozenset: _MyCustomList_set(a)
        if isset
        else (_MyCustomList_frozenset(a) if isfrozenset else a)
    )
    tree = jtu.tree_map(apply, tree, is_set, is_frozenset)

    _is_vnode = lambda x: isinstance(x, HDFVirtualNode)
    _is_leaf = lambda x: isinstance(x, HDFVirtualNode)

    with_v_nodes, without_v_nodes = eqx.partition(tree, _is_vnode, is_leaf=_is_leaf)
    leaves_with_path, struct = jtu.tree_flatten_with_path(with_v_nodes, is_leaf=_is_leaf)
    if len(leaves_with_path) == 0:
        logging.warning("No virtual nodes found.")
        return tree
    logging.info(f"Fetching {len(leaves_with_path)} leaf nodes from {type(tree).__name__}")

    assert all(isinstance(vn, HDFVirtualNode) for (_, vn) in leaves_with_path)

    filename = leaves_with_path[0][1].filename
    assert all(vn.filename == filename for (_, vn) in leaves_with_path), (
        f"Filenames of virtual nodes {tuple(vn.filename for (_, vn) in leaves_with_path)} do not match."
    )
    v_nodes_paths = [
        path_from_jax_keypath(p, getitem_transform=AbstractVxData.make_hdf_key) for (p, _) in leaves_with_path
    ]
    parents_paths = [vn._v_parent_path_seq for (_, vn) in leaves_with_path]
    attrs = [vn.key for (_, vn) in leaves_with_path]
    assert all(map(_match_child_parent_paths, v_nodes_paths, parents_paths)), (
        "Incompatible parent path and v_node path and parent path."
    )
    with tb.open_file(filename, mode="r") as hf5_file:
        fetched_nodes = []
        for (_, v_node), attr in zip(leaves_with_path, attrs):
            parent_group = hf5_file.get_node(v_node.parent_path)
            assert isinstance(parent_group, tb.Group)
            fetched = AbstractVxData.deserialize_object(
                parent_group, hdf_key=attr, attr_type_enum_name=v_node.type_enum, defer=(), levels=None
            )
            fetched_nodes.append(fetched)

        with_v_nodes = jtu.tree_unflatten(struct, fetched_nodes)
        tree = eqx.combine(without_v_nodes, with_v_nodes, is_leaf=_is_leaf)
        # return all transformed sets to lists back to sets (read Note 1 above).
        apply = (
            lambda a: set(a)
            if isinstance(a, _MyCustomList_set)
            else (frozenset(a) if isinstance(a, _MyCustomList_frozenset) else a)
        )
        return jtu.tree_map(apply, tree, is_leaf=lambda a: isinstance(a, (_MyCustomList_set, _MyCustomList_frozenset)))

    raise RuntimeError(f"Failed to open the HDFFile: {filename}.")



from abc import ABC, ABCMeta, abstractmethod
from collections.abc import Iterable, Generator
from functools import cached_property
from types import MappingProxyType
from typing import ClassVar, Self, Hashable

import equinox as eqx
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
import pandas as pd

from ._literals import SplitLiteral
from .base import AbstractConfig, AbstractVxData, fetch_at, HDFVirtualNode
from .coding_scheme import CodeMap, CodesVector, CodingSchemesManager, FilterOutcomeMap, GroupingData, ReducedCodeMapN1
from .dataset import (
    AbstractDatasetPipeline,
    AbstractProcessedDataset,
    AbstractTransformation,
    COLUMN,
    Dataset,
    DatasetSchemeConfig,
    DatasetSchemeProxy,
    PipelineReportTable,
    Report,
    ReportAttributes,
)
from .tvx_concepts import (
    Admission,
    AdmissionDates,
    DemographicVectorConfig,
    InpatientInput,
    InpatientInterventions,
    InpatientObservables,
    LeadingObservableExtractorConfig,
    Patient,
    SegmentedPatient,
    StaticInfo,
)
from .utils import Array


class ScalerConfig(AbstractConfig):
    use_float16: bool

    def __init__(self, use_float16: bool):
        self.use_float16 = use_float16


class TVxEHRSplitsConfig(AbstractConfig):
    split_quantiles: list[float]
    seed: int
    balance: SplitLiteral
    discount_first_admission: bool

    def __init__(
        self,
        split_quantiles: list[float],
        seed: int = 0,
        balance: SplitLiteral = "subjects",
        discount_first_admission: bool = False,
    ):
        self.split_quantiles = split_quantiles
        self.seed = seed
        self.balance = balance
        self.discount_first_admission = discount_first_admission


class TVxEHRSampleConfig(AbstractConfig):
    n_subjects: int
    seed: int
    offset: int

    def __init__(self, n_subjects: int, seed: int = 0, offset: int = 0):
        self.n_subjects = n_subjects
        self.seed = seed
        self.offset = offset


class IQROutlierRemoverConfig(AbstractConfig):
    outlier_q1: float
    outlier_q2: float
    outlier_iqr_scale: float
    outlier_z1: float
    outlier_z2: float

    def __init__(
        self,
        outlier_q1: float = 0.25,
        outlier_q2: float = 0.75,
        outlier_iqr_scale: float = 1.5,
        outlier_z1: float = -2.5,
        outlier_z2: float = 2.5,
    ):
        self.outlier_q1 = outlier_q1
        self.outlier_q2 = outlier_q2
        self.outlier_iqr_scale = outlier_iqr_scale
        self.outlier_z1 = outlier_z1
        self.outlier_z2 = outlier_z2


class OutlierRemoversConfig(AbstractConfig):
    obs: IQROutlierRemoverConfig | None

    def __init__(self, obs: IQROutlierRemoverConfig = IQROutlierRemoverConfig()):
        self.obs = obs


class ScalersConfig(AbstractConfig):
    obs: ScalerConfig | None
    icu_inputs: ScalerConfig | None

    def __init__(
        self,
        obs: ScalerConfig = ScalerConfig(use_float16=True),
        icu_inputs: ScalerConfig = ScalerConfig(use_float16=True),
    ):
        self.obs = obs
        self.icu_inputs = icu_inputs


class DatasetNumericalProcessorsConfig(AbstractConfig):
    scalers: ScalersConfig | None
    outlier_removers: OutlierRemoversConfig | None

    def __init__(
        self,
        scalers: ScalersConfig = ScalersConfig(),
        outlier_removers: OutlierRemoversConfig = OutlierRemoversConfig(),
    ):
        self.scalers = scalers
        self.outlier_removers = outlier_removers


class CodedValueProcessor(AbstractVxData, ABC):
    config: ScalerConfig | IQROutlierRemoverConfig
    table_name: str | None
    code_column: str | None
    value_column: str | None

    def __init__(
        self,
        config: ScalerConfig | IQROutlierRemoverConfig,
        table_name: str | None = None,
        code_column: str | None = None,
        value_column: str | None = None,
    ):
        self.config = config
        self.table_name = table_name
        self.code_column = code_column
        self.value_column = value_column

    def table_getter(self, dataset: Dataset) -> pd.DataFrame:
        assert self.table_name is not None, "Processor not fitted."
        return getattr(dataset.tables, self.table_name)

    def fit(
        self, dataset: Dataset, admission_ids: list[str], table_name: str, code_column: str, value_column: str
    ) -> Self:
        df = getattr(dataset.tables, table_name)
        c_adm_id = getattr(dataset.config.columns, table_name).admission_id
        df = df[[code_column, value_column, c_adm_id]]
        df = df[df[c_adm_id].isin(admission_ids)]

        fitted = self
        for k, v in self._extract_stats(df, code_column, value_column).items():
            fitted = eqx.tree_at(lambda x: getattr(x, k), fitted, v)

        for k, v in {"table_name": table_name, "code_column": code_column, "value_column": value_column}.items():
            fitted = eqx.tree_at(lambda x: getattr(x, k), fitted, v, is_leaf=lambda x: x is None)
        return fitted

    @abstractmethod
    def _extract_stats(self, df: pd.DataFrame, c_code: str, c_value: str) -> dict[str, pd.Series]:
        pass

    @abstractmethod
    def __call__(self, dataset: Dataset) -> Dataset:
        pass

    @property
    def series_dict(self) -> dict[str, pd.Series]:
        return {k: v for k, v in self.__dict__.items() if isinstance(v, pd.Series) and len(v) > 0}

    @property
    def processing_target(self) -> dict[str, str]:
        assert self.table_name is not None, "Processor not fitted."
        assert self.code_column is not None, "Processor not fitted."
        assert self.value_column is not None, "Processor not fitted."
        return {"table_name": self.table_name, "code_column": self.code_column, "value_column": self.value_column}


class CodedValueScaler(CodedValueProcessor, ABC):
    config: ScalerConfig

    @property
    @abstractmethod
    def original_dtype(self) -> np.dtype:
        pass

    @abstractmethod
    def unscale(self, array: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def unscale_code(self, array: np.ndarray, code_index: Hashable) -> np.ndarray:
        pass


def outcome_first_occurrence(sorted_admissions: list[Admission]) -> np.ndarray:
    """
    Find the first occurrence admission index of each outcome in a list of sorted admissions.

    Args:
        sorted_admissions (list[Admission]): a list of sorted Admission objects.

    Returns:
        np.ndarray: an array containing the admission index of the first occurrence of
        each outcome for each admission. -1 means no occurrence.
    """
    first_occurrence = np.empty_like(sorted_admissions[0].outcome.vec, dtype=type[sorted_admissions[0].admission_id])
    first_occurrence[:] = -1
    for adm in sorted_admissions:
        update_mask = (first_occurrence < 0) & adm.outcome.vec
        first_occurrence[update_mask] = adm.admission_id
    return first_occurrence


class OutlierRemovers(AbstractVxData):
    obs: CodedValueProcessor | None

    def __init__(self, obs: CodedValueProcessor | None = None):
        self.obs = obs


class Scalers(AbstractVxData):
    obs: CodedValueScaler | None
    icu_inputs: CodedValueScaler | None

    def __init__(self, obs: CodedValueScaler | None = None, icu_inputs: CodedValueScaler | None = None):
        self.obs = obs
        self.icu_inputs = icu_inputs


class DatasetNumericalProcessors(AbstractVxData):
    outlier_removers: OutlierRemovers
    scalers: Scalers

    def __init__(self, outlier_removers: OutlierRemovers = OutlierRemovers(), scalers: Scalers = Scalers()):
        self.outlier_removers = outlier_removers
        self.scalers = scalers


class TVxEHRSchemeConfig(DatasetSchemeConfig):
    outcome: str | None = None

    def __init__(
        self,
        ethnicity: str | None = None,
        gender: str | None = None,
        dx_discharge: str | None = None,
        obs: str | None = None,
        icu_procedures: str | None = None,
        hosp_procedures: str | None = None,
        icu_inputs: str | None = None,
        outcome: str | None = None,
    ):
        super().__init__(
            ethnicity=ethnicity,
            gender=gender,
            dx_discharge=dx_discharge,
            obs=obs,
            icu_procedures=icu_procedures,
            hosp_procedures=hosp_procedures,
            icu_inputs=icu_inputs,
        )
        self.outcome = outcome


class TVxEHRSchemeProxy(DatasetSchemeProxy):
    config: TVxEHRSchemeConfig

    def __init__(self, config: TVxEHRSchemeConfig, schemes_context: CodingSchemesManager):
        super().__init__(config=config, schemes_context=schemes_context)

    @cached_property
    def outcome(self) -> FilterOutcomeMap | None:
        if self.config.outcome is None or self.config.dx_discharge is None:
            return None
        return self.schemes_context.outcome[self.config.dx_discharge, self.config.outcome]

    @cached_property
    def outcome_size(self) -> int | None:
        return len(self.outcome) if self.outcome else None

    @cached_property
    def outcome_base_mapper(self) -> CodeMap | None:
        return self.outcome.codemap if self.outcome else None

    @staticmethod
    def validate_mapping(
        coding_scheme_manager: CodingSchemesManager, source: DatasetSchemeConfig, target: TVxEHRSchemeConfig
    ):
        target_schemes = target.scheme_fields()
        for key, source_scheme in source.scheme_fields().items():
            target_scheme = target_schemes[key]
            if target_scheme is None or source_scheme is None:
                continue
            assert coding_scheme_manager.map[(source_scheme, target_scheme)] is not None, (
                f"Cannot map {key} from {source_scheme} to {target_scheme}"
            )

        if target.outcome is not None:
            assert target.dx_discharge is not None, "Outcome specified but dx_discharge is None."
            assert coding_scheme_manager.supported_outcome(target.outcome, target.dx_discharge), (
                f"Outcome {target.outcome} not supported for {target.dx_discharge}"
            )

    def demographic_vector_size(self, demographic_vector_config: DemographicVectorConfig):
        size = 0
        if demographic_vector_config.gender:
            assert self.gender is not None, "Gender vector not initialized."
            size += len(self.gender)
        if demographic_vector_config.age:
            size += 1
        if demographic_vector_config.ethnicity:
            assert self.ethnicity is not None, "Ethnicity vector not initialized."
            size += len(self.ethnicity)
        return size

    def mapper(self, source_scheme_config: DatasetSchemeConfig, key: str) -> CodeMap | None:
        pair = (getattr(source_scheme_config, key), getattr(self, key).name)
        return self.schemes_context.map[pair] if pair in self.schemes_context.map else None

    def dx_mapper(self, source_scheme_config: DatasetSchemeConfig) -> CodeMap | None:
        return self.mapper(source_scheme_config, "dx_discharge")

    def ethnicity_mapper(self, source_scheme_config: DatasetSchemeConfig) -> CodeMap | None:
        return self.mapper(source_scheme_config, "ethnicity")

    def gender_mapper(self, source_scheme_config: DatasetSchemeConfig) -> CodeMap | None:
        return self.mapper(source_scheme_config, "gender")

    def icu_procedures_mapper(self, source_scheme_config: DatasetSchemeConfig) -> CodeMap | None:
        return self.mapper(source_scheme_config, "icu_procedures")

    def hosp_procedures_mapper(self, source_scheme_config: DatasetSchemeConfig) -> CodeMap | None:
        return self.mapper(source_scheme_config, "hosp_procedures")

    def icu_inputs_mapper(self, source_scheme_config: DatasetSchemeConfig) -> ReducedCodeMapN1 | None:
        m = self.mapper(source_scheme_config, "icu_inputs")
        assert isinstance(m, ReducedCodeMapN1) or m is None, "ICU inputs mapper must be a ReducedCodeMapN1 or None."
        return m

    def icu_inputs_grouping(self, source_scheme_config: DatasetSchemeConfig) -> GroupingData | None:
        mapper = self.icu_inputs_mapper(source_scheme_config)
        if mapper is None:
            return None
        assert source_scheme_config.icu_inputs is not None, "ICU inputs scheme not configured."
        return mapper.grouping_data(self.schemes_context.scheme[source_scheme_config.icu_inputs].index)

    @staticmethod
    def check_target_scheme_support(
        scheme_manager: CodingSchemesManager, dataset_scheme: DatasetSchemeProxy
    ) -> dict[str, tuple[str, ...]]:
        supported_attr_targets = {
            k: scheme_manager.scheme_supported_targets(v)
            for k, v in dataset_scheme.scheme_dict.items()
            if v is not None
        }
        assert dataset_scheme.dx_discharge is not None, "dx_discharge scheme not configured."
        supported_outcomes = scheme_manager.supported_outcomes(dataset_scheme.dx_discharge.name)
        return supported_attr_targets | {"outcome": supported_outcomes}


class TVxEHRConfig(AbstractConfig):
    scheme: TVxEHRSchemeConfig
    demographic: DemographicVectorConfig
    sample: TVxEHRSampleConfig | None
    splits: TVxEHRSplitsConfig | None
    numerical_processors: DatasetNumericalProcessorsConfig
    interventions: bool
    observables: bool
    time_binning: float | None
    leading_observable: LeadingObservableExtractorConfig | None
    interventions_segmentation: bool

    def __init__(
        self,
        scheme: TVxEHRSchemeConfig,
        demographic: DemographicVectorConfig,
        sample: TVxEHRSampleConfig | None = None,
        splits: TVxEHRSplitsConfig | None = None,
        numerical_processors: DatasetNumericalProcessorsConfig = DatasetNumericalProcessorsConfig(),
        interventions: bool = False,
        observables: bool = False,
        time_binning: float | None = None,
        leading_observable: LeadingObservableExtractorConfig | None = None,
        interventions_segmentation: bool = False,
    ):
        self.scheme = scheme
        self.demographic = demographic
        self.sample = sample
        self.splits = splits
        self.numerical_processors = numerical_processors
        self.interventions = interventions
        self.observables = observables
        self.time_binning = time_binning
        self.leading_observable = leading_observable
        self.interventions_segmentation = interventions_segmentation


class TVxReportAttributes(ReportAttributes):
    tvx_concept: str | None = None

    def __init__(self, tvx_concept: str | None = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tvx_concept = tvx_concept

    @staticmethod
    def _t(object_or_type: type | AbstractVxData) -> str:
        return object_or_type.__name__ if isinstance(object_or_type, type) else object_or_type.__class__.__name__

    @classmethod
    def ehr_prefix(cls) -> str:
        return cls._t(TVxEHR)

    @classmethod
    def subjects_prefix(cls) -> str:
        return f"{cls.ehr_prefix()}.dict[str, {cls._t(Patient)}]"

    @classmethod
    def subject_prefix(cls) -> str:
        return f"{cls.subjects_prefix()}[i]"

    @classmethod
    def static_info_prefix(cls) -> str:
        return f"{cls.subject_prefix()}.{cls._t(StaticInfo)}"

    @classmethod
    def admissions_prefix(cls) -> str:
        return f"{cls.subject_prefix()}.list[{cls._t(Admission)}]"

    @classmethod
    def inpatient_input_prefix(cls, attribute) -> str:
        return f"{cls.admissions_prefix()}[j].{cls._t(InpatientInterventions)}.{attribute}({cls._t(InpatientInput)})"

    @classmethod
    def admission_attribute_prefix(cls, attribute: str, attribute_type: str | type) -> str:
        if isinstance(attribute_type, type):
            attribute_type = cls._t(attribute_type)

        return f"{cls.admissions_prefix()}[j].{attribute}({attribute_type})"

    @classmethod
    def admission_codes_attribute(cls, attribute) -> str:
        return f"{cls.admission_attribute_prefix(attribute, CodesVector)}"


class TVxReport(Report):
    incident_class: ClassVar[type[TVxReportAttributes]] = TVxReportAttributes


class AbstractTVxPipeline(AbstractDatasetPipeline, metaclass=ABCMeta):
    transformations: list[AbstractTransformation]
    report_class: ClassVar[type[TVxReport]] = TVxReport


_SplitsType = tuple[tuple[str, ...], ...]


class TVxEHR(AbstractProcessedDataset):
    config: TVxEHRConfig
    dataset: Dataset
    numerical_processors: DatasetNumericalProcessors
    splits: _SplitsType | None
    subjects: MappingProxyType[str, Patient] | None
    patient_class: ClassVar[type[Patient]] = Patient
    report_class: ClassVar[type[TVxReport]] = TVxReport

    def __init__(
        self,
        config: TVxEHRConfig,
        dataset: Dataset,
        numerical_processors: DatasetNumericalProcessors = DatasetNumericalProcessors(),
        splits: _SplitsType | None = None,
        subjects: dict[str, Patient] | None = None,
        pipeline_report: PipelineReportTable = PipelineReportTable(),
    ):
        self.config = config
        self.dataset = dataset
        self.numerical_processors = numerical_processors
        self.splits = splits
        self.subjects = subjects  # type: ignore
        self.pipeline_report = PipelineReportTable(pipeline_report)

    @property
    def subject_ids(self) -> tuple[str, ...]:
        """Get the list of subject IDs."""
        return tuple(sorted(self.subjects.keys())) if self.subjects is not None else ()

    def fetch_subjects(self, subject_ids: tuple[str, ...] | None = None) -> Self:
        if subject_ids is None:
            subject_ids = self.subject_ids
        # generating lambdas inside generators can lead to unexpected behaviour, e.g. all lambdas can be bounded
        # to one value of subject_id (the last one of the sequence).
        # https://stackoverflow.com/a/452660
        return fetch_at(tuple(map(lambda k: lambda x: x.subjects[k], subject_ids)), self)

    def try_fetch_subjects(self, subject_ids: tuple[str, ...] | None = None) -> Self:
        if subject_ids is None:
            subject_ids = self.subject_ids

        # only fetch those not loaded already.
        assert self.subjects is not None, "Subjects not initialized."
        subject_ids = tuple(i for i in subject_ids if isinstance(self.subjects[i], HDFVirtualNode))
        return self.fetch_subjects(subject_ids)

    def scheme_proxy(self, schemes_context: CodingSchemesManager) -> TVxEHRSchemeProxy:
        """Get the scheme."""
        scheme = TVxEHRSchemeProxy(self.config.scheme, schemes_context)
        TVxEHRSchemeProxy.validate_mapping(schemes_context, self.dataset.config.scheme, self.config.scheme)
        return scheme

    def gender_mapper(self, schemes_proxy: TVxEHRSchemeProxy) -> CodeMap | None:
        return schemes_proxy.gender_mapper(self.dataset.config.scheme)

    def ethnicity_mapper(self, schemes_proxy: TVxEHRSchemeProxy) -> CodeMap | None:
        return schemes_proxy.ethnicity_mapper(self.dataset.config.scheme)

    def dx_mapper(self, schemes_proxy: TVxEHRSchemeProxy) -> CodeMap | None:
        return schemes_proxy.dx_mapper(self.dataset.config.scheme)

    def icu_procedures_mapper(self, schemes_proxy: TVxEHRSchemeProxy) -> CodeMap | None:
        return schemes_proxy.icu_procedures_mapper(self.dataset.config.scheme)

    def hosp_procedures_mapper(self, schemes_proxy: TVxEHRSchemeProxy) -> CodeMap | None:
        return schemes_proxy.hosp_procedures_mapper(self.dataset.config.scheme)

    def icu_inputs_grouping(self, schemes_proxy: TVxEHRSchemeProxy) -> GroupingData | None:
        return schemes_proxy.icu_inputs_grouping(self.dataset.config.scheme)

    def demographic_vector_size(self, schemes_proxy: TVxEHRSchemeProxy) -> int:
        return schemes_proxy.demographic_vector_size(self.config.demographic)

    @cached_property
    def subjects_sorted_admission_ids(self) -> dict[str, list[str]]:
        admissions = self.dataset.tables.admissions
        # For each subject, get the list of admission id sorted by admission date.
        sorted_index = lambda x: x.sort_values().index.to_list()
        return admissions.groupby(COLUMN.subject_id)[COLUMN.start_time].apply(sorted_index).to_dict()

    @cached_property
    def admission_ids(self) -> list[str]:
        admission_ids = sum(self.subjects_sorted_admission_ids.values(), [])
        # Check unique admission IDs across all subjects.
        if len(admission_ids) != len(set(admission_ids)):
            raise ValueError("Duplicate admission IDs found.")

        return admission_ids

    def subject_admission_demographics(self, subject_id: str) -> dict[str, Array]:
        assert self.subjects is not None, "Subjects not initialized."
        return self.subjects[subject_id].admission_demographics(self.config.demographic)

    @cached_property
    def admission_demographics(self) -> dict[str, Array]:
        assert self.subjects is not None, "Subjects not initialized."
        return {
            admission_id: admission_demo
            for subject_id in self.subjects
            for admission_id, admission_demo in self.subject_admission_demographics(subject_id).items()
        }

    @cached_property
    def admission_dates(self) -> dict[str, AdmissionDates]:
        admissions = self.dataset.tables.admissions
        c_admittime = self.dataset.config.columns.admissions.start_time
        c_dischtime = self.dataset.config.columns.admissions.end_time
        return admissions.apply(lambda x: AdmissionDates(x[c_admittime], x[c_dischtime]), axis=1).to_dict()  # type: ignore

    def fetch_device_batch(self, subject_ids: tuple[str, ...] | None = None) -> tuple[Self, Self]:
        # 1. Fetch from disk if the subjects are lazy-loaded in a new tree `ehr`.
        # 2. Load the subjects to the device in a new tree `device_ehr`.
        # 3. return as a tuple (`ehr`, `device_ehr`) as `device_ehr` can be re-used avoiding disk reading.
        if subject_ids is None:
            subject_ids = self.subject_ids
        ehr = self.try_fetch_subjects(subject_ids)

        assert ehr.subjects is not None, "Subjects not initialized."
        device_subjects = {
            i: ehr.subjects[i].to_jax_arrays()
            for i in subject_ids
            # tqdm_constructor(subject_ids,
            # desc="Loading to device",
            # unit='subject',
            # leave=False)
        }
        return ehr, eqx.tree_at(lambda x: x.subjects, self, device_subjects)

    def epoch_splits(
        self, subject_ids: Iterable[str] | None, batch_n_admissions: int, discount_first_admission: bool = False
    ) -> tuple[tuple[str, ...], ...]:
        """Generate epoch splits for training.

        Args:
            subject_ids (Optional[list[str]]): list of subject IDs to split.
            batch_n_admissions (int): number of admissions per batch.
            discount_first_admission (bool, optional): whether to ignore the first admission from the counts.
            Defaults to False.

        Returns:
            list[list[str]]: list of lists containing the split subject IDs.
        """

        n_adms = self.dataset.subjects_n_admissions
        if discount_first_admission:
            n_adms = n_adms - 1
        if subject_ids is None:
            subject_ids_arr = n_adms.index
        else:
            subject_ids_arr = n_adms.loc[list(subject_ids)].index

        n_splits = n_adms.loc[subject_ids_arr].sum() // batch_n_admissions
        if n_splits == 0:
            n_splits = 1
        p_splits = np.linspace(0, 1, n_splits + 1)[1:-1]

        w_adms = n_adms.loc[subject_ids_arr] / n_adms.loc[subject_ids_arr].sum()
        weights = w_adms.values.cumsum()
        splits = np.searchsorted(weights, p_splits)
        splits = (a.tolist() for a in np.split(subject_ids_arr, splits))
        return tuple(tuple(s) for s in splits if len(s) > 0)

    def iter_obs(self, subject_ids: Iterable[str] | None = None) -> Generator[InpatientObservables]:
        """Iterate over the observables for the given subject IDs.

        Args:
            subject_ids: list of subject IDs.

        Yields:
            InpatientObservables: InpatientObservables object.
        """
        assert self.subjects is not None, "Subjects not initialized."
        if subject_ids is None:
            subject_ids = self.subjects.keys()
        for s in subject_ids:
            for adm in self.subjects[s].admissions:
                assert adm.observables is not None, "Observables not initialized."
                yield adm.observables

    def iter_lead_obs(self, subject_ids: Iterable[str] | None = None) -> Generator[InpatientObservables]:
        """Iterate over the leading observables for the given subject IDs.

        Args:
            subject_ids: list of subject IDs.

        Yields:
            InpatientObservables: InpatientObservables object.
        """
        assert self.subjects is not None, "Subjects not initialized."
        if subject_ids is None:
            subject_ids = self.subjects.keys()
        for s in subject_ids:
            for adm in self.subjects[s].admissions:
                assert adm.leading_observable is not None, "Leading observable not initialized."
                yield adm.leading_observable

    def n_obs_times(self, subject_ids: Iterable[str] | None = None) -> int:
        """Get the total number of observation times.

        Args:
            subject_ids: list of subject IDs.

        Returns:
            int: total number of observation times.
        """
        return sum(len(obs) for obs in self.iter_obs(subject_ids))

    def d2d_interval_days(self, subject_ids: Iterable[str] | None = None) -> float:
        """Get the total number of days between first discharge and last discharge.

        Args:
            subject_ids: list of subject IDs.

        Returns:
            int: total number of days between first discharge and last discharge.
        """
        assert self.subjects is not None, "Subjects not initialized."
        if subject_ids is None:
            subject_ids = self.subjects.keys()

        return sum(self.subjects[s].d2d_interval_days for s in subject_ids)

    def interval_days(self, subject_ids: Iterable[str] | None = None) -> float:
        """Get the total number of days in-hospital.

        Args:
            subject_ids: list of subject IDs.

        Returns:
            int: total number of days in-hospital.
        """
        assert self.subjects is not None, "Subjects not initialized."
        if subject_ids is None:
            subject_ids = self.subjects.keys()

        return sum(a.interval_days for s in subject_ids for a in self.subjects[s].admissions)

    def interval_hours(self, subject_ids: Iterable[str] | None = None) -> float:
        """Get the total number of hours in-hospital.

        Args:
            subject_ids: list of subject IDs.

        Returns:
            int: Total number of hours in-hospital.
        """
        assert self.subjects is not None, "Subjects not initialized."
        if subject_ids is None:
            subject_ids = self.subjects.keys()

        return sum(a.interval_hours for s in subject_ids for a in self.subjects[s].admissions)

    def p_obs(self, subject_ids: Iterable[str] | None = None) -> float:
        """For a collection of subjects, compute a measure that is proportional to the rate of presence per
        observation timestamp.

        Args:
            subject_ids: list of subject IDs.

        Returns:
            float: proportion of observables presence per unique timestamp.
        """

        return float(sum(obs.mask.sum() for obs in self.iter_obs(subject_ids)) / self.n_obs_times())

    def obs_coocurrence_matrix(self, subject_ids: Iterable[str] | None = None) -> jnp.ndarray:
        """Compute the co-occurrence (or co-presence) matrix of observables.

        Returns:
            jnp.ndarray: co-occurrence (or co-presence) matrix of observables.
        """
        obs = [obs.mask for obs in self.iter_obs(subject_ids) if len(obs) > 0]
        obs = jnp.vstack(obs, dtype=int)
        return obs.T @ obs

    def size_in_bytes(self) -> int:
        """Get the size of the Patients object in bytes.

        Returns:
            int: size of the Patients object in bytes.
        """
        is_arr = eqx.filter(self.subjects, eqx.is_array)
        arr_size = jtu.tree_map(lambda a, m: a.size * a.itemsize if m is not None else 0, self.subjects, is_arr)
        return sum(jtu.tree_leaves(arr_size))

    def _unscaled_observation(self, obs: InpatientObservables) -> InpatientObservables:
        """Unscale the observation values, undo the preprocessing scaling.

        Args:
            obs (InpatientObservables): observation to unscale.

        Returns:
            InpatientObservables: unscaled observation.
        """
        obs_scaler = self.numerical_processors.scalers.obs
        assert obs_scaler is not None, "No observation scaler configured."
        value = obs_scaler.unscale(np.array(obs.value))
        return InpatientObservables(time=obs.time, value=value, mask=obs.mask)

    def _unscaled_leading_observable(self, lead: InpatientObservables, code_index: int) -> InpatientObservables:
        """Unscale the leading observable values, undo the preprocessing scaling.

        Args:
            lead (InpatientObservables): leading observable to unscale.

        Returns:
            InpatientObservables: unscaled leading observable.
        """
        lead_scaler = self.numerical_processors.scalers.obs
        assert lead_scaler is not None, "No observation scaler configured."
        value = lead_scaler.unscale_code(np.array(lead.value), code_index)
        return InpatientObservables(time=lead.time, value=value, mask=lead.mask)

    def subject_size_in_bytes(self, subject_id: str) -> int:
        """Get the size of the subject object in bytes.

        Args:
            subject_id (str): subject ID.

        Returns:
            int: size of the subject object in bytes.
        """
        assert self.subjects is not None, "Subjects not initialized."
        is_arr = eqx.filter(self.subjects[subject_id], eqx.is_array)
        arr_size = jtu.tree_map(
            lambda a, m: a.size * a.itemsize if m is not None else 0, self.subjects[subject_id], is_arr
        )
        return sum(jtu.tree_leaves(arr_size))

    def outcome_frequency_vec(self, subjects: Iterable[str]) -> Array:
        """Get the outcome frequency vector for a list of subjects.

        Args:
            subjects (list[str]): list of subject IDs.

        Returns:
            jnp.ndarray: outcome frequency vector.
        """
        assert self.subjects is not None, "Subjects not initialized."
        return sum(self.subjects[i].outcome_frequency_vec() for i in subjects)

    def outcome_frequency_partitions(self, n_partitions: int, subjects: Iterable[str]) -> tuple[tuple[int, ...], ...]:
        """
        Get the outcome codes partitioned by their frequency of occurrence into `n_partitions` partitions. The codes in
        each partition contribute to 1 / n_partitions of the all outcome occurrences.

        Args:
            n_partitions (int): number of partitions.
            subjects (list[str]): list of subject IDs.

        Returns:
            list[list[int]]: a list of outcome codes partitioned by frequency into `n_partitions` partitions.

        """
        frequency_vec = self.outcome_frequency_vec(subjects)
        frequency_vec = frequency_vec / frequency_vec.sum()
        sorted_codes = np.argsort(frequency_vec)
        frequency_vec = frequency_vec[sorted_codes]
        cumsum = np.cumsum(frequency_vec)
        partitions = np.linspace(0, 1, n_partitions + 1)[1:-1]
        splitters = np.searchsorted(cumsum, partitions)
        return tuple(tuple(p.astype(int)) for p in np.hsplit(sorted_codes, splitters))

    def outcome_first_occurrence(self, subject_id: str) -> np.ndarray:
        """Get the first occurrence admission index of each outcome for a subject. If an outcome does not occur,
        the index is set to -1.

        Args:
            subject_id (str): subject ID.

        Returns:
            int: first occurrence admission index of each outcome for a subject.
        """
        assert self.subjects is not None, "Subjects not initialized."
        return outcome_first_occurrence(self.subjects[subject_id].admissions)

    def outcome_first_occurrence_masks(self, subject_id: str) -> tuple[Array, ...]:
        """Get a list of masks indicating whether an outcome occurs for a subject for the first time.

        Args:
            subject_id (str): subject ID.

        Returns:
            list[bool]: list of masks indicating whether an outcome occurs for a subject for the first time.

        """
        assert self.subjects is not None, "Subjects not initialized."
        adms = self.subjects[subject_id].admissions
        first_occ_adm_id = outcome_first_occurrence(adms)
        return tuple(first_occ_adm_id == a.admission_id for a in adms)

    def outcome_all_masks(self, subject_id: str) -> tuple[Array, ...]:
        """Get a list of full-masks with the same shape as the outcome vector."""
        assert self.subjects is not None, "Subjects not initialized."
        adms = self.subjects[subject_id].admissions
        if isinstance(adms[0].outcome.vec, jnp.ndarray):
            _np = jnp
        else:
            _np = np
        return tuple(_np.ones_like(a.outcome.vec, dtype=bool) for a in adms)


class SegmentedTVxEHR(TVxEHR):
    subjects: dict[str, SegmentedPatient]
    patient_class: ClassVar[type[SegmentedPatient]] = SegmentedPatient

    def __init__(
        self,
        config: TVxEHRConfig,
        dataset: Dataset,
        numerical_processors: DatasetNumericalProcessors = DatasetNumericalProcessors(),
        splits: _SplitsType | None = None,
        subjects: dict[str, SegmentedPatient] | None = None,
        pipeline_report: PipelineReportTable = PipelineReportTable(),
    ):
        super().__init__(
            config=config,
            dataset=dataset,
            numerical_processors=numerical_processors,
            splits=splits,
            subjects=subjects,  # type: ignore
            pipeline_report=pipeline_report,
        )

    def execute_pipeline(
        self, pipeline: AbstractDatasetPipeline, schemes_context: CodingSchemesManager
    ) -> AbstractProcessedDataset:
        raise NotImplementedError("SegmentedPatient is a final representation. It cannot have a pipeline.")

    @classmethod
    def from_tvx_ehr(
        cls, tvx_ehr: TVxEHR, schemes_context: CodingSchemesManager, maximum_padding: int = 100
    ) -> "SegmentedTVxEHR":
        dataset_scheme_proxy = tvx_ehr.dataset.scheme_proxy(schemes_context)
        tvx_scheme_proxy = tvx_ehr.scheme_proxy(schemes_context)
        hosp_procedures_size = len(tvx_scheme_proxy.hosp_procedures) if tvx_scheme_proxy.hosp_procedures else None
        icu_procedures_size = len(tvx_scheme_proxy.icu_procedures) if tvx_scheme_proxy.icu_procedures else None
        icu_inputs_size = len(dataset_scheme_proxy.icu_inputs) if dataset_scheme_proxy.icu_inputs else None
        assert tvx_ehr.subjects is not None, "Subjects not initialized."
        subjects = {
            k: SegmentedPatient.from_patient(
                v,
                hosp_procedures_size=hosp_procedures_size,
                icu_procedures_size=icu_procedures_size,
                icu_inputs_size=icu_inputs_size,
                maximum_padding=maximum_padding,
            )
            for k, v in tvx_ehr.subjects.items()
        }
        return SegmentedTVxEHR(
            config=tvx_ehr.config,
            dataset=tvx_ehr.dataset,
            numerical_processors=tvx_ehr.numerical_processors,
            subjects=subjects,
            splits=tvx_ehr.splits,
            pipeline_report=tvx_ehr.pipeline_report,
        )



"""."""

import dataclasses
import enum
import logging
import random
from abc import ABC, ABCMeta, abstractmethod
from collections import defaultdict
from collections.abc import Iterator
from dataclasses import field
from datetime import datetime
from functools import cached_property
from typing import ClassVar, Final, Self

import equinox as eqx
import numpy as np
import pandas as pd

from ._literals import NumericalTypeHint, SplitLiteral
from ._stats.dataset import DatasetStatsInterface, MultiDatasetsStatsInterface, TwoDatasetsStatsInterface
from .base import AbstractConfig, AbstractVxData, HDFVirtualNode
from .coding_scheme import CodingScheme, CodingSchemesManager, NumericScheme
from .utils import tqdm_constructor

SECONDS_TO_HOURS_SCALER: Final[float] = 1 / 3600.0  # convert seconds to hours
SECONDS_TO_DAYS_SCALER: Final[float] = SECONDS_TO_HOURS_SCALER * 1 / 24.0  # convert seconds to days


# This Enum will be used as a reference to ensure consistent column names
# across the multiple (relational) columns representing a single dataset.
class COLUMN(enum.StrEnum):
    subject_id = enum.auto()
    admission_id = enum.auto()
    gender = enum.auto()
    race = enum.auto()
    date_of_birth = enum.auto()
    anchor_year = enum.auto()
    anchor_age = enum.auto()
    code = enum.auto()
    version = enum.auto()
    description = enum.auto()
    time = enum.auto()  # for singly timestamped events.
    measurement = enum.auto()
    start_time = enum.auto()  # for events defined by intervals rather than a single timestamp
    end_time = enum.auto()  # ..
    amount = enum.auto()
    amount_unit = enum.auto()
    derived_unit_normalization_factor = enum.auto()
    derived_universal_unit = enum.auto()
    derived_normalized_amount = enum.auto()
    derived_normalized_amount_per_hour = enum.auto()
    mapped_code = enum.auto()
    mapped_description = enum.auto()

    @property
    def is_time(self):
        return self in (
            COLUMN.time,
            COLUMN.start_time,
            COLUMN.end_time,
            COLUMN.date_of_birth,
            COLUMN.anchor_year,
        )

    @property
    def is_code(self):
        return self in (COLUMN.code, COLUMN.mapped_code)

    @property
    def is_id(self):
        return self in (COLUMN.subject_id, COLUMN.admission_id)

    @staticmethod
    def as_dict() -> dict[str, str]:
        """
        Returns a dictionary representation of the enum member.
        """
        return {str(c): c.value for c in COLUMN}


class TableColumns(AbstractConfig):
    def __check_init__(self):
        self.validate()

    def validate(self):
        assert all(k in COLUMN and k == v for k, v in self.as_dict().items()), f"Fields must be one of {COLUMN}."

    def as_tuple(self) -> tuple[str, ...]:
        return tuple(self.as_dict().values())

    @property
    def id_cols(self) -> tuple[str, ...]:
        return tuple(v for k, v in self.as_one_level_dict().items() if COLUMN[k].is_id)

    def __contains__(self, item: str | COLUMN):
        if isinstance(item, str):
            return item in self.as_one_level_dict()
        elif isinstance(item, COLUMN):
            return str(item) in self.as_one_level_dict()
        else:
            raise ValueError(f"Unsupported type {type(item)}")

    @property
    def index(self) -> tuple[str, ...]:
        return tuple(f.name for f in dataclasses.fields(self) if f.metadata.get("index", False))

    @property
    def time_cols(self) -> tuple[str, ...]:
        return tuple(v for k, v in self.as_dict().items() if COLUMN[k].is_time)

    @property
    def code_cols(self) -> tuple[str, ...]:
        return tuple(v for k, v in self.as_dict().items() if COLUMN[k].is_code)


class StaticTableColumns(TableColumns):
    subject_id: str = field(default=str(COLUMN.subject_id), metadata={"index": True})
    race: str = str(COLUMN.race)
    gender: str = str(COLUMN.gender)
    date_of_birth: str = str(COLUMN.date_of_birth)


class AdmissionsTableColumns(TableColumns):
    subject_id: str = str(COLUMN.subject_id)
    admission_id: str = field(default=str(COLUMN.admission_id), metadata={"index": True})
    start_time: str = str(COLUMN.start_time)
    end_time: str = str(COLUMN.end_time)


class AdmissionSummaryTableColumns(TableColumns):
    admission_id: str = str(COLUMN.admission_id)
    code: str = str(COLUMN.code)
    description: str = str(COLUMN.description)


class AdmissionTimeSeriesTableColumns(TableColumns):
    admission_id: str = str(COLUMN.admission_id)
    code: str = str(COLUMN.code)
    time: str = str(COLUMN.time)
    measurement: str = str(COLUMN.measurement)
    description: str = str(COLUMN.description)


class AdmissionIntervalEventsTableColumns(TableColumns):
    admission_id: str = str(COLUMN.admission_id)
    code: str = str(COLUMN.code)
    description: str = str(COLUMN.description)
    start_time: str = str(COLUMN.start_time)
    end_time: str = str(COLUMN.end_time)


class AdmissionIntervalRatesTableColumns(TableColumns):
    admission_id: str = str(COLUMN.admission_id)
    code: str = str(COLUMN.code)
    description: str = str(COLUMN.description)
    start_time: str = str(COLUMN.start_time)
    end_time: str = str(COLUMN.end_time)
    amount: str = str(COLUMN.amount)
    amount_unit: str = str(COLUMN.amount_unit)
    derived_unit_normalization_factor: str = str(COLUMN.derived_unit_normalization_factor)
    derived_universal_unit: str = str(COLUMN.derived_universal_unit)
    derived_normalized_amount: str = str(COLUMN.derived_normalized_amount)
    derived_normalized_amount_per_hour: str = str(COLUMN.derived_normalized_amount_per_hour)


class MultivariateTimeSeriesTableMeta(AbstractConfig):
    name: str
    attributes: tuple[str, ...]
    type_hint: tuple[NumericalTypeHint, ...]
    default_type_hint: NumericalTypeHint

    def __init__(
        self,
        name: str,
        attributes: tuple[str, ...],
        type_hint: tuple[NumericalTypeHint, ...] | None = None,
        default_type_hint: NumericalTypeHint = "N",
    ):
        self.name = name
        self.attributes = attributes
        self.default_type_hint = default_type_hint
        self.type_hint = type_hint or ((default_type_hint,) * len(attributes))

    def __check_init__(self):
        assert len(self.attributes) == len(self.type_hint), (
            f"Length of attributes and type_hint must be the same. Got {len(self.attributes)} and "
            f"{len(self.type_hint)}."
        )
        assert all(t in ("N", "C", "B", "O") for t in self.type_hint), (
            f"type hint must be one of 'N', 'C', 'B', 'O'. Got {self.type_hint}."
        )


class DatasetColumns(AbstractConfig):
    static: StaticTableColumns
    admissions: AdmissionsTableColumns
    dx_discharge: AdmissionSummaryTableColumns
    obs: AdmissionTimeSeriesTableColumns
    icu_procedures: AdmissionIntervalEventsTableColumns
    icu_inputs: AdmissionIntervalRatesTableColumns
    hosp_procedures: AdmissionIntervalEventsTableColumns

    def __init__(
        self,
        static: StaticTableColumns = StaticTableColumns(),
        admissions: AdmissionsTableColumns = AdmissionsTableColumns(),
        dx_discharge: AdmissionSummaryTableColumns = AdmissionSummaryTableColumns(),
        obs: AdmissionTimeSeriesTableColumns = AdmissionTimeSeriesTableColumns(),
        icu_procedures: AdmissionIntervalEventsTableColumns = AdmissionIntervalEventsTableColumns(),
        icu_inputs: AdmissionIntervalRatesTableColumns = AdmissionIntervalRatesTableColumns(),
        hosp_procedures: AdmissionIntervalEventsTableColumns = AdmissionIntervalEventsTableColumns(),
    ):
        self.static = static
        self.admissions = admissions
        self.dx_discharge = dx_discharge
        self.obs = obs
        self.icu_procedures = icu_procedures
        self.icu_inputs = icu_inputs
        self.hosp_procedures = hosp_procedures

    def __check_init__(self):
        self.validate()

    def validate(self):
        assert all(isinstance(v, TableColumns) for v in self.as_one_level_dict().values())
        for v in self.as_one_level_dict().values():
            v.validate()
        column_names = defaultdict(set)
        for v in self.as_dict().values():
            for k, v in v.items():
                assert k == v
                column_names[k].add(v)
        for k, v in column_names.items():
            if len(v) > 1:
                raise ValueError(f"Column {k} is present with different names: {v}")

    def columns_dict(self) -> dict[str, tuple[str, ...]]:
        return {k: v.as_dict() for k, v in self.as_one_level_dict().items()}

    @property
    def admission_id(self) -> str:
        return self.admissions.admission_id

    @property
    def subject_id(self) -> str:
        return self.static.subject_id

    @property
    def timestamped_tables_config_dict(self):
        return {k: v for k, v in self.as_one_level_dict().items() if COLUMN.time in v.as_dict().keys()}

    @property
    def interval_based_table_config_dict(self):
        return {
            k: v
            for k, v in self.as_one_level_dict().items()
            if {COLUMN.start_time, str(COLUMN.end_time)}.issubset(set(v.as_dict().keys()))
        }

    @property
    def indices(self) -> dict[str, str]:
        return {k: v.index for k, v in self.as_one_level_dict().items() if len(v.index) > 0}

    @property
    def time_cols(self) -> dict[str, tuple[str, ...]]:
        return {k: v.time_cols for k, v in self.as_one_level_dict().items() if len(v.time_cols) > 0}

    @property
    def code_column(self) -> dict[str, str]:
        return {k: v.code_cols for k, v in self.as_one_level_dict().items() if len(v.code_cols) > 0}

    def temporal_admission_linked_table(self, table_name: str) -> bool:
        conf = getattr(self, table_name)
        return len(conf.time_cols) > 0 and COLUMN.admission_id in conf and str(COLUMN.admission_id) not in conf.index


class DatasetTables(AbstractVxData):
    static: pd.DataFrame
    admissions: pd.DataFrame
    dx_discharge: pd.DataFrame | None
    obs: pd.DataFrame | None
    icu_procedures: pd.DataFrame | None
    icu_inputs: pd.DataFrame | None
    hosp_procedures: pd.DataFrame | None

    def __init__(
        self,
        static: pd.DataFrame,
        admissions: pd.DataFrame,
        dx_discharge: pd.DataFrame | None = None,
        obs: pd.DataFrame | None = None,
        icu_procedures: pd.DataFrame | None = None,
        icu_inputs: pd.DataFrame | None = None,
        hosp_procedures: pd.DataFrame | None = None,
    ):
        self.static = static
        self.admissions = admissions
        self.dx_discharge = dx_discharge
        self.obs = obs
        self.icu_procedures = icu_procedures
        self.icu_inputs = icu_inputs
        self.hosp_procedures = hosp_procedures

    def __check_init__(self):
        if isinstance(self.admissions, HDFVirtualNode):
            return
        if COLUMN.admission_id in self.admissions.columns:
            admission_id = self.admissions[COLUMN.admission_id]
        elif self.admissions.index.name == str(COLUMN.admission_id):
            admission_id = self.admissions.index
        else:
            raise ValueError(
                f"Where is the admission_id? columns: {self.admissions.columns}. Index: {self.admissions.index.name}"
            )

        assert admission_id.nunique() == len(admission_id), (
            "Admission IDs in MIMIC-III and MIMIC-IV were found to be globally unique, i.e. two patients cannot share "
            "an admission ID. This allowed simpler dataset representation. Since we have the admissions table listing "
            "both subject_id and admission_id (unique 1:1 relation), no need to include the subject_id in other tables."
            "However, in the future there might be a need for an extension/adaptation to deal with non-unique "
            "admission IDs, where both (subject_id, admission_id) are needed for identification. "
            "In case you are getting this error message, you can either rewrite all admission_ids of your dataset "
            "tables to be globally unique, or, if not an urgent request, you can post an Issue "
            "at the repository of this code or email at: (asem.a.abdelaziz@proton.me). "
            "TODO: fix this potential limitation. Refer to this issue in the comments with ISSUE_ADM_UNIQ"
        )

    @property
    def tables_dict(self) -> dict[str, pd.DataFrame]:
        return {k: v for k, v in self.__dict__.items() if isinstance(v, pd.DataFrame)}


class DatasetSchemeConfig(AbstractConfig):
    ethnicity: str | None
    gender: str | None
    dx_discharge: str | None
    obs: str | None
    icu_procedures: str | None
    hosp_procedures: str | None
    icu_inputs: str | None

    def __init__(
        self,
        ethnicity: str | None = None,
        gender: str | None = None,
        dx_discharge: str | None = None,
        obs: str | None = None,
        icu_procedures: str | None = None,
        hosp_procedures: str | None = None,
        icu_inputs: str | None = None,
    ):
        self.ethnicity = ethnicity
        self.gender = gender
        self.dx_discharge = dx_discharge
        self.obs = obs
        self.icu_procedures = icu_procedures
        self.hosp_procedures = hosp_procedures
        self.icu_inputs = icu_inputs

    def scheme_fields(self) -> dict[str, str | None]:
        return {
            "gender": self.gender,
            "ethnicity": self.ethnicity,
            "dx_discharge": self.dx_discharge,
            "obs": self.obs,
            "icu_inputs": self.icu_inputs,
            "icu_procedures": self.icu_procedures,
            "hosp_procedures": self.hosp_procedures,
        }


@dataclasses.dataclass
class DatasetSchemeProxy:
    """
    Represents a dataset scheme that defines the coding schemes and outcome extractor for a dataset.

    Attributes:
        config (DatasetSchemeConfig): the configuration for the dataset scheme.

    Methods:
        __init__(self, config: DatasetSchemeConfig, **kwargs): initializes a new instance of the DatasetScheme class.
        scheme_dict(self): returns a dictionary of the coding schemes in the dataset scheme.
        make_target_scheme_config(self, **kwargs): creates a new target scheme configuration based on the
            current scheme.
        make_target_scheme(self, config=None, **kwargs): creates a new target scheme based on the current scheme.
        demographic_vector_size(self, demographic_vector_config: DemographicVectorConfig): calculates the size of
            the demographic vector.
        dx_mapper(self, target_scheme: DatasetScheme): returns the mapper for the diagnosis coding scheme to the
            corresponding target scheme.
        ethnicity_mapper(self, target_scheme: DatasetScheme): returns the mapper for the ethnicity coding scheme to
            the corresponding target scheme.
        supported_target_scheme_options(self): returns the supported target scheme options for each coding scheme.
    """

    config: DatasetSchemeConfig
    schemes_context: CodingSchemesManager

    def __init__(self, config: DatasetSchemeConfig, schemes_context: CodingSchemesManager):
        self.config = config
        self.schemes_context = schemes_context

    def _scheme(self, name: str | None) -> CodingScheme | None:
        if name is None:
            return None
        try:
            return self.schemes_context.scheme[name]
        except KeyError:
            return None

    @property
    def ethnicity(self) -> CodingScheme | None:
        return self._scheme(self.config.ethnicity)

    @property
    def gender(self) -> CodingScheme | None:
        return self._scheme(self.config.gender)

    @property
    def dx_discharge(self) -> CodingScheme | None:
        return self._scheme(self.config.dx_discharge)

    @property
    def obs(self) -> NumericScheme | None:
        s = self._scheme(self.config.obs)
        assert s is None or isinstance(s, NumericScheme), f"Observation scheme must be numeric. Got {type(s)}."
        return s

    @property
    def icu_procedures(self) -> CodingScheme | None:
        return self._scheme(self.config.icu_procedures)

    @property
    def hosp_procedures(self) -> CodingScheme | None:
        return self._scheme(self.config.hosp_procedures)

    @property
    def icu_inputs(self) -> CodingScheme | None:
        s = self._scheme(self.config.icu_inputs)
        return s

    @property
    def scheme_dict(self):
        return {k: self._scheme(v) for k, v in self.config.scheme_fields().items() if self._scheme(v) is not None}


class ReportAttributes(AbstractConfig):
    transformation: str | None = None
    operation: str | None = None
    table: str | None = None
    column: str | None = None
    value_type: str | None = None
    before: str | int | float | bool | None = None
    after: str | int | float | bool | None = None
    timestamp: str | None = None

    def __init__(
        self,
        transformation: str | type | None = None,
        operation: str | None = None,
        table: str | None = None,
        column: str | None = None,
        value_type: str | type | None = None,
        before: str | int | float | bool | type | None = None,
        after: str | int | float | bool | type | None = None,
        timestamp: str | None = None,
    ):
        def strtype(x):
            if x is None:
                return x
            elif isinstance(x, type):
                return x.__name__
            elif isinstance(x, np.dtype):
                return x.name
            else:
                return x

        self.transformation = strtype(transformation)
        self.operation = operation
        self.table = table
        self.column = column
        self.value_type = strtype(value_type)
        self.before = strtype(before)
        self.after = strtype(after)
        if timestamp is not None:
            self.timestamp = timestamp
        else:
            self.timestamp = datetime.now().isoformat()


class PipelineReportTable(pd.DataFrame):
    # We need to exclude the timestamps of the steps from the equality tests.
    def equals(self, other: object) -> bool:  # type: ignore[override]
        assert isinstance(other, pd.DataFrame), f"Can only compare with pd.DataFrame. Got {type(other)}."
        # Exclude timestamps from comparison.
        report = self
        if all("timestamp" in r for r in (self.columns, other.columns)):
            report = report.drop(columns=["timestamp"])
            other = other.drop(columns=["timestamp"])
        return pd.DataFrame.equals(report, other)


class Report(AbstractConfig):
    incidents: tuple[ReportAttributes, ...]
    incident_class: ClassVar[type[ReportAttributes]] = ReportAttributes

    def __init__(self, incidents: tuple[ReportAttributes, ...] = ()):
        self.incidents = incidents

    def __add__(self, other: Self) -> Self:
        return type(self)(incidents=self.incidents + other.incidents)

    def add(self, *args, **kwargs) -> Self:
        return type(self)(incidents=self.incidents + (self.incident_class(*args, **kwargs),))

    def __getitem__(self, item) -> ReportAttributes:
        return self.incidents[item]

    def __iter__(self) -> Iterator[ReportAttributes]:
        return iter(self.incidents)

    def __len__(self) -> int:
        return len(self.incidents)

    def compile(self, previous_report: pd.DataFrame | None = None) -> PipelineReportTable:
        report = self.incidents
        if len(report) == 0:
            report = (ReportAttributes(transformation="identity"),)

        df = pd.DataFrame([x.as_dict() for x in report]).astype(str)
        object_columns = [c for c in df.columns if df[c].dtype == "object"]
        type_rows = df["value_type"] == "dtype"
        type_cols = ["after", "before"]
        nan_mask = df.loc[:, object_columns].isnull() | df.loc[:, object_columns].isin((None, "nan", "NaN", "None"))
        df.loc[:, object_columns] = df.loc[:, object_columns].where(~nan_mask, "-")
        df.loc[type_rows, type_cols] = df.loc[type_rows, type_cols].map(lambda x: f"{x}_type")
        if previous_report is None:
            return PipelineReportTable(df)
        else:
            return PipelineReportTable(pd.concat([previous_report, df], ignore_index=True, axis=0, sort=False))


class AbstractDataset(AbstractVxData, ABC):
    config: AbstractConfig
    report_class: ClassVar[type[Report]] = Report

    @abstractmethod
    def scheme_proxy(self, schemes_context: CodingSchemesManager) -> DatasetSchemeProxy: ...


class AbstractTransformation[DType: AbstractDataset, RType: Report](eqx.Module):
    @classmethod
    @abstractmethod
    def apply(cls, dataset: DType, schemes_context: CodingSchemesManager, report: RType, /) -> tuple[DType, RType]:
        raise NotImplementedError

    @classmethod
    def skip(cls, dataset: DType, report: RType, reason: str = "") -> tuple[DType, RType]:
        return dataset, report.add(transformation=cls, operation=": ".join(("skip", reason)))


class AbstractDatasetPipelineConfig(AbstractConfig):
    pass


class AbstractDatasetPipeline(AbstractVxData, metaclass=ABCMeta):
    config: AbstractDatasetPipelineConfig
    transformations: list[AbstractTransformation]
    report_class: ClassVar[type[Report]] = Report

    def __init__(
        self,
        config: AbstractDatasetPipelineConfig = AbstractDatasetPipelineConfig(),
        *,
        transformations: list[AbstractTransformation],
    ):
        self.config = config
        self.transformations = transformations


class AbstractProcessedDataset(AbstractDataset):
    pipeline_report: PipelineReportTable

    @property
    def pipeline_executed(self) -> bool:
        return len(self.pipeline_report) != 0

    def execute_pipeline(self, pipeline: AbstractDatasetPipeline, schemes_context: CodingSchemesManager) -> Self:
        if self.pipeline_executed:
            logging.warning("A pipeline has already been executed. Doing nothing.")
            return self
        return self._execute_pipeline(pipeline.transformations, schemes_context)

    def _execute_pipeline(
        self, transformations: list[AbstractTransformation], schemes_context: CodingSchemesManager
    ) -> Self:
        if self.pipeline_executed:
            logging.warning("A pipeline has already been executed. This will replace the pipeline execution history.")
        report = self.report_class()
        dataset = self
        with tqdm_constructor(desc="Transforming Dataset", unit="transformations", total=len(transformations)) as pbar:
            for t in transformations:
                pbar.set_description(f"Transforming Dataset: {type(t).__name__}")
                report = report.add(transformation=type(t), operation="start")
                dataset, report = t.apply(dataset, schemes_context, report)
                report = report.add(transformation=type(t), operation="end")
                pbar.update(1)

        return eqx.tree_at(lambda x: x.pipeline_report, dataset, report.compile(dataset.pipeline_report))


class DatasetConfig(AbstractConfig):
    scheme: DatasetSchemeConfig
    columns: DatasetColumns
    select_subjects_with_observation: str | None
    select_subjects_with_short_admissions: float | None  # number of days.
    admission_minimum_los: float | None

    def __init__(
        self,
        scheme: DatasetSchemeConfig,
        columns: DatasetColumns = DatasetColumns(),
        select_subjects_with_observation: str | None = None,
        select_subjects_with_short_admissions: float | None = None,
        admission_minimum_los: float | None = None,
    ):
        self.scheme = scheme
        self.columns = columns
        self.select_subjects_with_observation = select_subjects_with_observation
        self.select_subjects_with_short_admissions = select_subjects_with_short_admissions
        self.admission_minimum_los = admission_minimum_los


class Dataset(AbstractProcessedDataset):
    """
    A class representing a dataset.

    Attributes:
        config (DatasetConfig): the configuration object for the dataset.

    Methods:
        __init__(self, config: DatasetConfig = None, config_path: str = None, **kwargs): initializes the Dataset object.
        supported_target_scheme_options(self): returns the supported target scheme options.
        to_subjects(self, **kwargs): converts the dataset to subject objects.
        save(self, path: Union[str, Path], overwrite: bool = False): saves the dataset to disk.
        load(cls, path: Union[str, Path]): loads the dataset from disk.
    """

    config: DatasetConfig
    tables: DatasetTables

    def __init__(
        self, config: DatasetConfig, tables: DatasetTables, pipeline_report: PipelineReportTable = PipelineReportTable()
    ):
        self.config = config
        self.tables = tables
        self.pipeline_report = PipelineReportTable(pipeline_report)

    def scheme_proxy(self, schemes_context: CodingSchemesManager) -> DatasetSchemeProxy:
        return DatasetSchemeProxy(self.config.scheme, schemes_context)

    def stats(self, coding_schemes_manager: CodingSchemesManager) -> DatasetStatsInterface:
        return DatasetStatsInterface(self, coding_schemes_manager)

    @classmethod
    def multi_stats(cls, *datasets: Self, coding_schemes_manager: CodingSchemesManager) -> MultiDatasetsStatsInterface:
        return MultiDatasetsStatsInterface(*datasets, schemes_manager=coding_schemes_manager)

    @classmethod
    def two_stats(
        cls, dataset1: Self, dataset2: Self, coding_schemes_manager: CodingSchemesManager
    ) -> TwoDatasetsStatsInterface:
        return TwoDatasetsStatsInterface(dataset1, dataset2, schemes_manager=coding_schemes_manager)

    @cached_property
    def subject_ids(self) -> pd.Index:
        assert self.tables.static.index.name == self.config.columns.static.subject_id, (
            f"Index name of static table must be {self.config.columns.static.subject_id}."
        )
        return self.tables.static.index.unique()

    @cached_property
    def subjects_intervals_sum(self) -> pd.Series:
        c_admittime = self.config.columns.admissions.start_time
        c_dischtime = self.config.columns.admissions.end_time
        c_subject_id = self.config.columns.admissions.subject_id
        admissions = self.tables.admissions
        interval = (admissions[c_dischtime] - admissions[c_admittime]).dt.total_seconds()
        admissions = admissions.assign(interval=interval)
        missed_subjects = set(self.subject_ids).difference(set(admissions[c_subject_id]))
        intervals_sum = admissions.groupby(c_subject_id)["interval"].sum()
        assert isinstance(intervals_sum, pd.Series)
        missing_subjects = pd.Series([0] * len(missed_subjects), index=pd.Series(list(missed_subjects)))
        return pd.concat([intervals_sum, missing_subjects])

    @cached_property
    def subjects_n_admissions(self) -> pd.Series:
        c_subject_id = self.config.columns.admissions.subject_id
        admissions = self.tables.admissions
        missed_subjects = set(self.subject_ids).difference(set(admissions[c_subject_id]))
        n_admissions = [admissions.groupby(c_subject_id).size()]
        if len(missed_subjects) > 0:
            n_admissions.append(pd.Series([0] * len(missed_subjects), index=pd.Series(list(missed_subjects))))
        n_admissions_s = pd.concat(n_admissions)
        assert isinstance(n_admissions_s, pd.Series)
        return n_admissions_s

    def random_splits(
        self,
        splits: list[float],
        subject_ids: list[str] | None = None,
        random_seed: int = 42,
        balance: SplitLiteral = "subjects",
        discount_first_admission: bool = False,
    ) -> tuple[list[str], ...]:
        assert len(splits) > 0, "Split quantiles must be non-empty."
        assert list(splits) == sorted(splits), "Splits must be sorted."
        assert balance in ("subjects", "admissions", "admissions_intervals"), (
            "Balanced must be'subjects', 'admissions', or 'admissions_intervals'."
        )
        if subject_ids is None:
            subject_ids = list(self.subject_ids)
        assert len(subject_ids) > 0, "No subjects in the dataset."

        subject_ids = sorted(subject_ids)

        random.Random(random_seed).shuffle(subject_ids)
        subject_ids_arr = np.array(subject_ids)

        c_subject_id = self.config.columns.static.subject_id

        admissions = self.tables.admissions.loc[self.tables.admissions.loc[:, c_subject_id].isin(subject_ids_arr), :]

        if balance == "subjects":
            probs = (np.ones(len(subject_ids_arr)) / len(subject_ids_arr)).cumsum()

        elif balance == "admissions":
            assert len(admissions) > 0, "No admissions in the dataset."
            n_admissions = self.subjects_n_admissions.loc[subject_ids_arr]
            if discount_first_admission:
                n_admissions = n_admissions - 1
            p_admissions = n_admissions / n_admissions.sum()
            probs = p_admissions.values.cumsum()

        elif balance == "admissions_intervals":
            assert len(admissions) > 0, "No admissions in the dataset."
            subjects_intervals_sum = self.subjects_intervals_sum.loc[subject_ids_arr]
            p_subject_intervals = subjects_intervals_sum / subjects_intervals_sum.sum()
            probs = p_subject_intervals.values.cumsum()
        else:
            raise ValueError(f"Unknown balanced option: {balance}")

        # Deal with edge cases where the splits are exactly the same as the probabilities.
        for i in range(len(splits)):
            if any(abs(probs - splits[i]) < 1e-6):
                splits[i] = splits[i] + 1e-6

        splits_array = np.searchsorted(probs, splits)
        return tuple(a.tolist() for a in np.split(subject_ids_arr, splits_array))



from typing import Literal


CompressionLibLiteral = Literal["blosc", "zlib", "lzo", "bzip2"]
NumericalTypeHint = Literal["B", "N", "O", "C"]  # Binary, Numerical, Ordinal, Categorical
AggregationLiteral = Literal["sum", "or", "w_sum"]
SplitLiteral = Literal["subjects", "admissions", "admissions_intervals"]
TableAggregationLiteral = Literal["admission", "first_admission", "subject"]
